#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Functions to support paraxial ray tracing a sequential optical model

.. Created on Tue Feb 13 10:48:19 2018

.. codeauthor: Michael J. Hayford
"""
import math
from collections import namedtuple
from rayoptics.optical.model_constants import Intfc, Gap, Tfrm, Indx, Zdir
from rayoptics.optical.model_constants import ht, slp, aoi
from rayoptics.optical.idealimager import ideal_imager_setup

ParaxData = namedtuple('ParaxData', ['ax_ray', 'pr_ray', 'fod'])
""" tuple grouping together paraxial rays and first order properties

    Attributes:
        ax_ray: axial marginal ray data, y, u, i
        pr_ray: chief ray data, y, u, i
        fod: instance of :class:`~.FirstOrderData`
"""


class FirstOrderData:
    """ Container class for first order optical properties

    All quantities are based on paraxial ray tracing. The last interface is
    the image-1 interface.

    Attributes:
        opt_inv: optical invariant
        efl: effective focal length
        pp1: distance of front principle plane from 1st interface
        ppk: distance of rear principle plane from last interface
        ffl: front focal length
        bfl: back focal length
        fno: focal ratio at working conjugates, f/#
        red: reduction ratio
        n_obj: refractive index at central wavelength in object space
        n_img: refractive index at central wavelength in image space
        obj_dist: object distance
        img_dist: image distance
        obj_ang: maximum object angle (degrees)
        img_ht: image height
        enp_dist: entrance pupil distance from 1st interface
        enp_radius: entrance pupil radius
        exp_dist: exit pupil distance from last interface
        exp_radius: exit pupil radius
        obj_na: numerical aperture in object space
        img_na: numerical aperture in image space
    """

    def __init__(self):
        self.opt_inv = None
        self.power = None
        self.efl = None
        self.pp1 = None
        self.ppk = None
        self.ffl = None
        self.bfl = None
        self.fno = None
        self.m = None
        self.red = None
        self.n_obj = None
        self.n_img = None
        self.obj_dist = None
        self.img_dist = None
        self.obj_ang = None
        self.img_ht = None
        self.enp_dist = None
        self.enp_radius = None
        self.exp_dist = None
        self.exp_radius = None
        self.obj_na = None
        self.img_na = None

    def list_first_order_data(self):
        """ list the first order properties """
        print("efl        {:12.4g}".format(self.efl))
        print("ffl        {:12.4g}".format(self.ffl))
        print("pp1        {:12.4g}".format(self.pp1))
        print("bfl        {:12.4g}".format(self.bfl))
        print("ppk        {:12.4g}".format(self.ppk))
        print("f/#        {:12.4g}".format(self.fno))
        print("m          {:12.4g}".format(self.m))
        print("red        {:12.4g}".format(self.red))
        print("obj_dist   {:12.4g}".format(self.obj_dist))
        print("obj_ang    {:12.4g}".format(self.obj_ang))
        print("enp_dist   {:12.4g}".format(self.enp_dist))
        print("enp_radius {:12.4g}".format(self.enp_radius))
        print("na obj     {:12.4g}".format(self.obj_na))
        print("n obj      {:12.4g}".format(self.n_obj))
        print("img_dist   {:12.4g}".format(self.img_dist))
        print("img_ht     {:12.4g}".format(self.img_ht))
        print("exp_dist   {:12.4g}".format(self.exp_dist))
        print("exp_radius {:12.4g}".format(self.exp_radius))
        print("na img     {:12.4g}".format(self.img_na))
        print("n img      {:12.4g}".format(self.n_img))
        print("optical invariant {:12.4g}".format(self.opt_inv))


# paraxial_trace() - This routine performs a paraxial raytrace from object
#                    (surface 0) to image.
def paraxial_trace(path, start, start_yu, start_yu_bar):
    """ perform a paraxial raytrace of 2 linearly independent rays """
    p_ray = []
    p_ray_bar = []

    before = next(path)
    z_dir_before = before[Zdir]
    n_before = before[Indx] if z_dir_before > 0 else -before[Indx]

    b4_yui = start_yu
    b4_yui_bar = start_yu_bar
    if start == 1:
        # compute object coords from 1st surface data
        t0 = before[Gap].thi
        obj_ht = start_yu[ht] - t0*start_yu[slp]
        obj_htb = start_yu_bar[ht] - t0*start_yu_bar[slp]
        b4_yui = [obj_ht, start_yu[slp]]
        b4_yui_bar = [obj_htb, start_yu_bar[slp]]

    cv = before[Intfc].profile_cv
    # calculate angle of incidence (aoi)
    aoi = b4_yui[slp] + b4_yui[ht] * cv
    aoi_bar = b4_yui_bar[slp] + b4_yui_bar[ht] * cv
    b4_yui.append(aoi)
    b4_yui_bar.append(aoi_bar)

    p_ray.append(b4_yui)
    p_ray_bar.append(b4_yui_bar)

    # loop over remaining surfaces in path
    while True:
        try:
            after = next(path)

            z_dir_after = after[Zdir]
            n_after = after[Indx] if z_dir_after > 0 else -after[Indx]

            # Transfer
            t = before[Gap].thi
            cur_ht = b4_yui[ht] + t * b4_yui[slp]
            cur_htb = b4_yui_bar[ht] + t * b4_yui_bar[slp]

            # Refraction/Reflection
            srf = after[Intfc]
            k = n_before/n_after

            # calculate slope after refraction/reflection
            pwr = srf.optical_power
            cur_slp = k * b4_yui[slp] - cur_ht * pwr/n_after
            cur_slpb = k * b4_yui_bar[slp] - cur_htb * pwr/n_after

            # calculate angle of incidence (aoi)
            cv = srf.profile_cv
            aoi = cur_slp + cur_ht * cv
            aoi_bar = cur_slpb + cur_htb * cv

            yu = [cur_ht, cur_slp, aoi]
            yu_bar = [cur_htb, cur_slpb, aoi_bar]

            p_ray.append(yu)
            p_ray_bar.append(yu_bar)

            b4_yui = yu
            b4_yui_bar = yu_bar

            n_before = n_after
            z_dir_before = z_dir_after
            before = after

        except StopIteration:
            break

    return p_ray, p_ray_bar


def compute_first_order(opt_model, stop, wvl):
    """ Returns paraxial axial and chief rays, plus first order data. """
    seq_model = opt_model.seq_model
    start = 1
    n_0 = seq_model.z_dir[start-1]*seq_model.central_rndx(start-1)
    uq0 = 1/n_0
    p_ray, q_ray = paraxial_trace(seq_model.path(wl=wvl), start,
                                  [1., 0.], [0., uq0])

    n_k = seq_model.z_dir[-1]*seq_model.central_rndx(-1)
    img = -2 if seq_model.get_num_surfaces() > 2 else -1
    ak1 = p_ray[img][ht]
    bk1 = q_ray[img][ht]
    ck1 = n_k*p_ray[img][slp]
    dk1 = n_k*q_ray[img][slp]

    # print(p_ray[-2][ht], q_ray[-2][ht], n_k*p_ray[-2][slp], n_k*q_ray[-2][slp])
    # print(ak1, bk1, ck1, dk1)

    if stop is not None:
        n_s = seq_model.z_dir[stop]*seq_model.central_rndx(stop)
        as1 = p_ray[stop][ht]
        bs1 = q_ray[stop][ht]
        cs1 = n_s*p_ray[stop][slp]
        ds1 = n_s*q_ray[stop][slp]

        # find entrance pupil location w.r.t. first surface
        ybar1 = -bs1
        ubar1 = as1
        n_0 = seq_model.gaps[0].medium.rindex(wvl)
        enp_dist = -ybar1/(n_0*ubar1)

        thi0 = seq_model.gaps[0].thi

        # calculate reduction ratio for given object distance
        red = dk1 + thi0*ck1
        obj2enp_dist = thi0 + enp_dist

        yu = [0., 1.]
        pupil = opt_model.optical_spec.pupil
        aperture, obj_img_key, value_key = pupil.key
        if obj_img_key == 'object':
            if value_key == 'pupil':
                slp0 = 0.5*pupil.value/obj2enp_dist
            elif value_key == 'NA':
                slp0 = n_0*math.tan(math.asin(pupil.value/n_0))
        elif obj_img_key == 'image':
            if value_key == 'f/#':
                slpk = -1./(2.0*pupil.value)
                slp0 = slpk/red
            elif value_key == 'NA':
                slpk = n_k*math.tan(math.asin(pupil.value/n_k))
                slp0 = slpk/red
        yu = [0., slp0]

        yu_bar = [1., 0.]
        fov = opt_model.optical_spec.field_of_view
        field, obj_img_key, value_key = fov.key
        max_fld, fn = fov.max_field()
        if max_fld == 0.0:
            max_fld = 1.0
        if obj_img_key == 'object':
            if value_key == 'angle':
                ang = math.radians(max_fld)
                slpbar0 = math.tan(ang)
                ybar0 = -slpbar0*obj2enp_dist
            elif value_key == 'height':
                ybar0 = -max_fld
                slpbar0 = -ybar0/obj2enp_dist
        elif obj_img_key == 'image':
            if value_key == 'height':
                ybar0 = red*max_fld
                slpbar0 = -ybar0/obj2enp_dist
        yu_bar = [ybar0, slpbar0]

    else:  # floating stop surface - use parax_model for starting data
        ax = opt_model.parax_model.ax
        pr = opt_model.parax_model.pr
        yu = [0., ax[0][slp]/n_0]
        yu_bar = [pr[0][ht], pr[0][slp]/n_0]

    ax_ray, pr_ray = paraxial_trace(seq_model.path(wl=wvl), 0, yu, yu_bar)

    n_0 = seq_model.central_rndx(0)
    opt_inv = n_0*(ax_ray[1][ht]*pr_ray[0][slp] - pr_ray[1][ht]*ax_ray[0][slp])

    fod = FirstOrderData()
    fod.opt_inv = opt_inv
    fod.obj_dist = obj_dist = seq_model.gaps[0].thi
    fod.img_dist = img_dist = seq_model.gaps[-1].thi
    if ck1 == 0.0:
        fod.power = 0.0
        fod.efl = 0.0
        fod.pp1 = 0.0
        fod.ppk = 0.0
    else:
        fod.power = -ck1
        fod.efl = -1.0/ck1
        fod.pp1 = (dk1 - 1.0)*(n_0/ck1)
        fod.ppk = (p_ray[-2][ht] - 1.0)*(n_k/ck1)
    fod.ffl = fod.pp1 - fod.efl
    fod.bfl = fod.efl - fod.ppk
    fod.fno = -1.0/(2.0*n_k*ax_ray[-1][slp])

    fod.m = ak1 + ck1*img_dist/n_k
    fod.red = dk1 + ck1*obj_dist
    fod.n_obj = n_0
    fod.n_img = n_k
    fod.img_ht = -fod.opt_inv/(n_k*ax_ray[-1][slp])
    fod.obj_ang = math.degrees(math.atan(pr_ray[0][slp]))
    if pr_ray[0][slp] != 0:
        nu_pr0 = n_0*pr_ray[0][slp]
        fod.enp_dist = -pr_ray[1][ht]/nu_pr0
        fod.enp_radius = abs(fod.opt_inv/nu_pr0)
    else:
        fod.enp_dist = -1e10
        fod.enp_radius = 1e10

    if pr_ray[-1][slp] != 0:
        fod.exp_dist = -(pr_ray[-1][ht]/pr_ray[-1][slp] - fod.img_dist)
        fod.exp_radius = abs(fod.opt_inv/(n_k*pr_ray[-1][slp]))
    else:
        fod.exp_dist = -1e10
        fod.exp_radius = 1e10

    # compute object and image space numerical apertures
    fod.obj_na = n_0*math.sin(math.atan(seq_model.z_dir[0]*ax_ray[0][slp]))
    fod.img_na = n_k*math.sin(math.atan(seq_model.z_dir[-1]*ax_ray[-1][slp]))

    return ParaxData(ax_ray, pr_ray, fod)


def compute_principle_points(path, n_0=1.0, n_k=1.0):
    """ Returns paraxial p and q rays, plus partial first order data. """
    uq0 = 1/n_0
    p_ray, q_ray = paraxial_trace(path, 0, [1., 0.], [0., uq0])

    img = -1
    ak1 = p_ray[img][ht]
    bk1 = q_ray[img][ht]
    ck1 = n_k*p_ray[img][slp]
    dk1 = n_k*q_ray[img][slp]

    # print(p_ray[-2][ht], q_ray[-2][ht], n_k*p_ray[-2][slp], n_k*q_ray[-2][slp])
    # print(ak1, bk1, ck1, dk1)

    if ck1 == 0.0:
        efl = 0.0
        pp1 = 0.0
        ppk = 0.0
    else:
        efl = -1.0/ck1
        pp1 = (dk1 - 1.0)*(n_0/ck1)
        ppk = (p_ray[-1][ht] - 1.0)*(n_k/ck1)
    ffl = pp1 - efl
    bfl = efl - ppk

    return p_ray, q_ray, (efl, pp1, ppk, ffl, bfl)


def list_parax_trace(opt_model):
    """ list the paraxial axial and chief ray data """
    seq_model = opt_model.seq_model
    ax_ray, pr_ray, fod = opt_model.optical_spec.parax_data
    print("stop surface:", seq_model.stop_surface)
    print("           y           u           n*i         ybar         ubar"
          "        n*ibar")
    for i, ax in enumerate(ax_ray):
        n = seq_model.central_rndx(i)
        n = n if seq_model.z_dir[i] > 0 else -n
        print("{:2} {:12.6g} {:12.6g} {:12.6g} {:12.6g} {:12.6g} {:12.6g}"
              .format(i, ax_ray[i][ht], ax_ray[i][slp], n*ax_ray[i][aoi],
                      pr_ray[i][ht], pr_ray[i][slp], n*pr_ray[i][aoi]))


def specsheet_from_parax_data(opt_model, specsheet):
    """ update specsheet to contents of opt_model, while preserving inputs """
    if opt_model is None:
        return None
    seq_model = opt_model.seq_model
    optical_spec = opt_model.optical_spec
    fod = optical_spec.parax_data.fod
    conj_type = 'finite'
    if seq_model.gaps[0].thi > 10e8:
        conj_type = 'infinite'

    specsheet.conjugate_type = conj_type

    # specsheet.imager_inputs contains values of independent variables of
    # the optical system. Augment these as needed to get a defined imager.
    imager_inputs = dict(specsheet.imager_inputs)
    num_imager_inputs = len(imager_inputs)
    if num_imager_inputs == 0:
        # no user inputs, use model values
        if conj_type == 'finite':
            imager_inputs['m'] = fod.m
            imager_inputs['f'] = fod.efl
            specsheet.frozen_imager_inputs = [False]*5
        else:  # conj_type == 'infinite'
            imager_inputs['s'] = -math.inf
            if fod.efl != 0:
                imager_inputs['f'] = fod.efl
            specsheet.frozen_imager_inputs = [True, True, True, True, False]
    elif num_imager_inputs == 1:
        # some/partial user input specification
        if conj_type == 'finite':
            # make sure that m is specified
            if 'm' in imager_inputs:
                imager_inputs['f'] = fod.efl
            else:
                imager_inputs['m'] = fod.m
            specsheet.frozen_imager_inputs = [False]*5
        else:  # conj_type == 'infinite'
            imager_inputs['s'] = -math.inf
            if fod.efl != 0:
                imager_inputs['f'] = fod.efl
            specsheet.frozen_imager_inputs = [True, True, True, True, False]

    specsheet.imager = ideal_imager_setup(**imager_inputs)

    ape_key, ape_value = optical_spec.pupil.get_input_for_specsheet()
    fld_key, fld_value = optical_spec.field_of_view.get_input_for_specsheet()

    etendue_inputs = specsheet.etendue_inputs
    etendue_inputs[ape_key[0]][ape_key[1]][ape_key[2]] = ape_value
    etendue_inputs[fld_key[0]][fld_key[1]][fld_key[2]] = fld_value
    specsheet.generate_from_inputs(imager_inputs, etendue_inputs)

    return specsheet



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
"""thirder order aberration calculation

.. Created on Fri Jul  6 07:24:40 2018

.. codeauthor: Michael J. Hayford
"""

import numpy as np
import pandas as pd

from rayoptics.optical.model_constants import ht, slp, aoi


def compute_third_order(opt_model):
    """ Compute Seidel aberration coefficents. """
    seq_model = opt_model.seq_model
    n_before = seq_model.central_rndx(0)
    parax_data = opt_model.optical_spec.parax_data
    ax_ray, pr_ray, fod = parax_data
    opt_inv = fod.opt_inv
    opt_inv_sqr = opt_inv*opt_inv

    third_order = {}

    # Transfer from object
    p = 0
    pd_index = ['S-I', 'S-II', 'S-III', 'S-IV', 'S-V']
    for c in range(1, len(ax_ray)-1):
        n_after = seq_model.central_rndx(c)
        n_after = n_after if seq_model.z_dir[c] > 0 else -n_after
        cv = seq_model.ifcs[c].profile_cv

        A = n_after * ax_ray[c][aoi]
        Abar = n_after * pr_ray[c][aoi]
        P = cv*(1./n_after - 1./n_before)
        delta_slp = ax_ray[c][slp]/n_after - ax_ray[p][slp]/n_before
        SIi = -A**2 * ax_ray[c][ht] * delta_slp
        SIIi = -A*Abar * ax_ray[c][ht] * delta_slp
        SIIIi = -Abar**2 * ax_ray[c][ht] * delta_slp
        SIVi = -opt_inv_sqr * P
        delta_n_sqr = 1./n_after**2 - 1./n_before**2
        SVi = -Abar*(Abar * Abar * delta_n_sqr * ax_ray[c][ht] -
                     (opt_inv + Abar * ax_ray[c][ht])*pr_ray[c][ht]*P)

        scoef = pd.Series([SIi, SIIi, SIIIi, SIVi, SVi], index=pd_index)
        col = str(c)
        third_order[col] = scoef

        # handle case of aspheric profile
        if hasattr(seq_model.ifcs[c], 'profile'):
            to_asp = aspheric_seidel_contribution(seq_model, parax_data, c,
                                                  n_before, n_after)
            if to_asp:
                ascoef = pd.Series(to_asp, index=pd_index)
                third_order[col+'.asp'] = ascoef

        p = c
        n_before = n_after

    third_order_df = pd.DataFrame(third_order, index=pd_index)
    third_order_df['sum'] = third_order_df.sum(axis='columns')
    return third_order_df.T


def calc_4th_order_aspheric_term(p):
    G = 0.
    if type(p).__name__ == 'Conic':
        cv = p.cv
        cc = p.cc
        G = cc*cv**3/8.0
    elif type(p).__name__ == 'EvenPolynomial':
        cv = p.cv
        cc = p.cc
        G = cc*cv**3/8.0 + p.coef4

    return G


def aspheric_seidel_contribution(seq_model, parax_data, i, n_before, n_after):
    def delta_E(z, y, u, n):
        return -z/(n*y*(y + z*u))
    ax_ray, pr_ray, fod = parax_data

    if pr_ray[i][slp] == 0:
        e = pr_ray[i][ht]/ax_ray[i][ht]
    else:
        z = -pr_ray[i][ht]/pr_ray[i][slp]
        e = fod.opt_inv*delta_E(z, ax_ray[i][ht], ax_ray[i][slp], n_after)
    G = calc_4th_order_aspheric_term(seq_model.ifcs[i].profile)
    if G == 0.0:
        return None
    delta_n = n_after - n_before
    SI_star = 8.0*G*delta_n*ax_ray[i][ht]**4
    SII_star = SI_star*e
    SIII_star = SI_star*e**2
    SV_star = SI_star*e**3
    return [SI_star, SII_star, SIII_star, 0., SV_star]


def seidel_to_wavefront(seidel, central_wvl):
    """ Convert Seidel coefficients to wavefront aberrations """
    pd_index = ['W040', 'W131', 'W222', 'W220', 'W311']
    SI, SII, SIII, SIV, SV = seidel
    W040 = 0.125*SI/central_wvl
    W131 = 0.5*SII/central_wvl
    W222 = 0.5*SIII/central_wvl
    W220 = 0.25*(SIV + SIII)/central_wvl
    W311 = 0.5*SV/central_wvl
    return pd.Series([W040, W131, W222, W220, W311], index=pd_index)


def seidel_to_transverse_aberration(seidel, ref_index, slope):
    """ Convert Seidel coefficients to transverse ray aberrations """
    pd_index = ['TSA', 'TCO', 'TAS', 'SAS', 'PTB', 'DST']
    SI, SII, SIII, SIV, SV = seidel
    cnvrt = 1.0/(2.0*ref_index*slope)
    # TSA = transverse spherical aberration
    TSA = cnvrt*SI
    # TCO = tangential coma
    TCO = cnvrt*3.0*SII
    # TAS = tangential astigmatism
    TAS = cnvrt*(3.0*SIII + SIV)
    # SAS = sagittal astigmatism
    SAS = cnvrt*(SIII + SIV)
    # PTB = Petzval blur
    PTB = cnvrt*SIV
    # DST = distortion
    DST = cnvrt*SV
    return pd.Series([TSA, TCO, TAS, SAS, PTB, DST], index=pd_index)


def seidel_to_field_curv(seidel, ref_index, opt_inv):
    """ Convert Seidel coefficients to astigmatic and Petzval curvatures """
    pd_index = ['TCV', 'SCV', 'PCV']
    SI, SII, SIII, SIV, SV = seidel
    cnvrt = ref_index/opt_inv**2
    # TCV = curvature of the tangential image surface
    TCV = cnvrt*(3.0*SIII + SIV)
    # SCV = curvature of the sagittal image surface
    SCV = cnvrt*(SIII + SIV)
    # PCV = curvature of the Petzval surface
    PCV = cnvrt*SIV
    return pd.Series([TCV, SCV, PCV], index=pd_index)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" optical model constants

.. Created on Wed May 23 16:00:55 2018

.. codeauthor: Michael J. Hayford
"""

# sequential ray trace data structures
Intfc, Gap, Tfrm, Indx, Zdir = range(5)

Surf, Gap = range(2)

# paraxial optics data structures
# 4 parts of paraxial model: axial ray, principal ray,
#                            lens data, optical invariant
ax, pr, lns, inv = range(4)
# paraxial ray data at an interface: height, n*slope, n*angle of incidence
ht, slp, aoi = range(3)
# lens data: power, reduced distance, refractive index (n), refract mode
pwr, tau, indx, rmd = range(4)

# ray trace segment data
# p: intersection point with interface
# d: direction cosine exiting the interface
# dst: distance from intersection point to next interface
# nrml: surface normal at intersection point
# phase: optical phase introduced at the interface
p, d, dst, nrml, phase = range(5)

# ray package
# ray: list of ray segment data
# op:  optical path wrt equally inclined chords to the optical axis
# wvl: wavelength (in nm) that the ray was traced in
ray, op, wvl = range(3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
"""
.. Created on Wed Apr 18 11:04:53 2018

.. codeauthor: Michael J. Hayford
"""

import numpy as np


def line(p1, p2):
    A = (p1[1] - p2[1])
    B = (p2[0] - p1[0])
    C = (p1[0]*p2[1] - p2[0]*p1[1])
    return A, B, -C


def intersection(L1, L2):
    D = L1[0] * L2[1] - L1[1] * L2[0]
    Dx = L1[2] * L2[1] - L1[1] * L2[2]
    Dy = L1[0] * L2[2] - L1[2] * L2[0]
    if D != 0:
        x = Dx / D
        y = Dy / D
        return x, y
    else:
        return False


def get_intersect(a1, a2, b1, b2):
    """
    Returns the point of intersection of the lines passing through a2,a1 and b2,b1.
    a1: [x, y] a point on the first line
    a2: [x, y] another point on the first line
    b1: [x, y] a point on the second line
    b2: [x, y] another point on the second line
    """
    s = np.vstack([a1, a2, b1, b2])      # s for stacked
    h = np.hstack((s, np.ones((4, 1))))  # h for homogeneous
    l1 = np.cross(h[0], h[1])            # get first line
    l2 = np.cross(h[2], h[3])            # get second line
    x, y, z = np.cross(l1, l2)           # point of intersection
    if z == 0:                           # lines are parallel
        return (float('inf'), float('inf'))
    return (x/z, y/z)


def do_intersect(a1, a2, b1, b2, soln, delta):
    xy = get_intersect(a1, a2, b1, b2)
    d = xy - a1
    if abs(d[0]/delta[0] - d[1]/delta[1]) > 1e-14:
        print("t", d[0]/delta[0], d[1]/delta[1])
    if delta[0] != 0.0:
        t = d[0]/delta[0]
    else:
        t = d[1]/delta[1]
    if soln[0] < t and t < 1.0:
        soln = t, xy
    return soln


def intersect_with_3lines(pt, wht, bg, gr, rb):
#    print("wht_dist", wht_dist)
    soln = 0.0, pt
    delta = wht - pt
#    print("bg")
    soln = do_intersect(pt, wht, *bg, soln, delta)
#    print("gr")
    soln = do_intersect(pt, wht, *gr, soln, delta)
#    print("rb")
    soln = do_intersect(pt, wht, *rb, soln, delta)
    return soln[1]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Container class for optical usage information

.. Created on Thu Jan 25 11:01:04 2018

.. codeauthor: Michael J. Hayford
"""

import math
import numpy as np

from rayoptics.optical.firstorder import compute_first_order, list_parax_trace
from rayoptics.optical.trace import aim_chief_ray
from rayoptics.optical import model_enums
import rayoptics.optical.model_constants as mc
import rayoptics.util.colour_system as cs
from rayoptics.util import colors
srgb = cs.cs_srgb


class OpticalSpecs:
    """ Container class for optical usage information

    Contains optical usage information to specify the aperture, field of view,
    spectrum and focal position.

    It maintains a repository of paraxial data.

    Attributes:
        spectral_region: instance of :class:`~.WvlSpec`
        pupil: instance of :class:`~.PupilSpec`
        field_of_view: instance of :class:`~.FieldSpec`
        defocus: instance of :class:`~.FocusRange`
        parax_data: tuple of :obj:`~.firstorder.ParaxData`
    """

    do_aiming_default = True

    def __init__(self, opt_model, specsheet=None):
        self.opt_model = opt_model
        self.spectral_region = WvlSpec()
        self.pupil = PupilSpec(self)
        self.field_of_view = FieldSpec(self)
        self.defocus = FocusRange(0.0)
        self.parax_data = None
        self.do_aiming = OpticalSpecs.do_aiming_default
        if specsheet:
            self.set_from_specsheet(specsheet)

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['opt_model']
        del attrs['parax_data']
        del attrs['do_aiming']
        return attrs

    def set_from_list(self, dl):
        self.spectral_region = dl[0]
        self.pupil = dl[1]
        self.field_of_view = dl[2]

    def set_from_specsheet(self, ss):
        self.spectral_region.set_from_specsheet(ss)
        self.pupil.set_from_specsheet(ss)
        self.field_of_view.set_from_specsheet(ss)
        self.defocus.set_from_specsheet(ss)

    def sync_to_restore(self, opt_model):
        self.opt_model = opt_model
        if not hasattr(self, 'defocus'):
            self.defocus = FocusRange(0.0)
        if not hasattr(self, 'do_aiming'):
            self.do_aiming = OpticalSpecs.do_aiming_default

        self.spectral_region.sync_to_restore(self)
        self.pupil.sync_to_restore(self)
        self.field_of_view.sync_to_restore(self)

    def update_model(self):
        self.pupil.update_model()
        self.field_of_view.update_model()
        stop = self.opt_model.seq_model.stop_surface
        wvl = self.spectral_region.central_wvl

        self.parax_data = compute_first_order(self.opt_model, stop, wvl)
        if self.do_aiming and self.opt_model.seq_model.get_num_surfaces() > 2:
            for i, fld in enumerate(self.field_of_view.fields):
                aim_pt = aim_chief_ray(self.opt_model, fld, wvl)
                fld.aim_pt = aim_pt

    def lookup_fld_wvl_focus(self, fi, wl=None, fr=0.0):
        """ returns field, wavelength and defocus data

        Args:
            fi (int): index into the field_of_view list of Fields
            wl (int): index into the spectral_region list of wavelengths
            fr (float): focus range parameter, -1.0 to 1.0

        Returns:
            (**fld**, **wvl**, **foc**)

            - **fld** - :class:`Field` instance for field_of_view[fi]
            - **wvl** - wavelength in nm
            - **foc** - focus shift from image interface
        """
        if wl is None:
            wvl = self.spectral_region.central_wvl
        else:
            wvl = self.spectral_region.wavelengths[wl]
        fld = self.field_of_view.fields[fi]
        foc = self.defocus.get_focus(fr)
        return fld, wvl, foc

    def obj_coords(self, fld):
        fov = self.field_of_view
        fod = self.parax_data.fod
        field, obj_img_key, value_key = fov.key
        if obj_img_key == 'object':
            if value_key == 'angle':
                ang_dg = np.array([fld.x, fld.y, 0.0])
                dir_tan = np.tan(np.deg2rad(ang_dg))
                obj_pt = -dir_tan*(fod.obj_dist+fod.enp_dist)
            elif value_key == 'height':
                obj_pt = np.array([fld.x, fld.y, 0.0])
        elif obj_img_key == 'image':
            if value_key == 'height':
                img_pt = np.array([fld.x, fld.y, 0.0])
                obj_pt = fod.red*img_pt
        return obj_pt

    def list_first_order_data(self):
        self.parax_data.fod.list_first_order_data()

    def list_parax_trace(self):
        list_parax_trace(self.opt_model)


class WvlSpec:
    """ Class defining a spectral region

    A spectral region is a list of wavelengths (in nm) and corresponding
    weights. A reference wavelength index defines the "center" of the
    spectral region.

    """

    def __init__(self, wlwts=[(550., 1.)], ref_wl=0):
        self.set_from_list(wlwts)
        self.reference_wvl = ref_wl
        self.coating_wvl = 550.0

    @property
    def central_wvl(self):
        return self.wavelengths[self.reference_wvl]

    @central_wvl.setter
    def central_wvl(self, wvl):
        self.wavelengths[self.reference_wvl] = wvl

    def set_from_list(self, wlwts):
        self.wavelengths = []
        self.spectral_wts = []
        for wlwt in wlwts:
            self.wavelengths.append(wlwt[0])
            self.spectral_wts.append(wlwt[1])
        self.calc_colors()

    def sync_to_restore(self, optical_spec):
        self.calc_colors()

    def set_from_specsheet(self, ss):
        pass

    def add(self, wl, wt):
        self.wavelengths.append(wl)
        self.spectral_wts.append(wt)
        self.spectrum.sort(key=lambda w: w[0], reverse=True)

    def calc_colors(self):
        accent = colors.accent_colors()
        self.render_colors = []
        num_wvls = len(self.wavelengths)
        if num_wvls == 1:
            self.render_colors.append(accent['green'])
        elif num_wvls == 2:
            self.render_colors.append(accent['red'])
            self.render_colors.append(accent['blue'])
        elif num_wvls == 3:
            self.render_colors.append(accent['red'])
            self.render_colors.append(accent['green'])
            self.render_colors.append(accent['blue'])
        else:
            for w in self.wavelengths:
                print("calc_colors", w)
                rgb = srgb.wvl_to_rgb(w)
                print("rgb", rgb)
                self.render_colors.append(rgb)


class PupilSpec:
    """ Aperture specification

    Attributes:
        pupil_type: :class:`~.PupilType` enum
        value: size of the pupil
        pupil_rays: list of relative pupil coordinates for pupil limiting rays
        ray_labels: list of string labels for pupil_rays
    """
    default_pupil_rays = [[0., 0.], [1., 0.], [-1., 0.], [0., 1.], [0., -1.]]
    default_ray_labels = ['00', '+X', '-X', '+Y', '-Y']

    def __init__(self, parent, key=('object', 'pupil'), value=1.0):
        self.optical_spec = parent
        self.key = 'aperture', key[0], key[1]
        self.value = value
        self.pupil_rays = PupilSpec.default_pupil_rays
        self.ray_labels = PupilSpec.default_ray_labels

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['optical_spec']
        return attrs

    def sync_to_restore(self, optical_spec):
        if hasattr(self, 'pupil_type'):
            self.key = model_enums.get_ape_key_for_type(self.pupil_type)
            del self.pupil_type
        self.optical_spec = optical_spec

    def set_from_list(self, ppl_spec):
        self.key = model_enums.get_ape_key_for_type(ppl_spec[0])
        self.value = ppl_spec[1]

    def set_from_specsheet(self, ss):
        for k, v in ss.etendue_inputs['aperture'].items():
            if len(v) > 0:
                obj_img_key = k
                for k1, v1 in v.items():
                    value_key = k1
                    break

        self.key = 'aperture', obj_img_key, value_key
        self.value = ss.etendue_inputs['aperture'][obj_img_key][value_key]

    def get_input_for_specsheet(self):
        return self.key, self.value

    def update_model(self):
        if not hasattr(self, 'pupil_rays'):
            self.pupil_rays = PupilSpec.default_pupil_rays
            self.ray_labels = PupilSpec.default_ray_labels

    def get_pupil_type(self):
        return model_enums.get_ape_type_for_key(self.key).value

    def mutate_pupil_type(self, new_pupil_type):
        ape_key = model_enums.get_ape_key_for_type(new_pupil_type)
        aperture, obj_img_key, value_key = ape_key
        if self.optical_spec is not None:
            if self.optical_spec.parax_data is not None:
                fod = self.optical_spec.parax_data.fod
                if obj_img_key == 'object':
                    if value_key == 'pupil':
                        self.value = 2*fod.enp_radius
                    elif value_key == 'NA':
                        self.value = fod.obj_na
                elif obj_img_key == 'image':
                    if value_key == 'f/#':
                        self.value = fod.fno
                    elif value_key == 'NA':
                        self.value = fod.img_na

        self.key = ape_key


class FieldSpec:
    """ Field of view specification

    Attributes:
        field_type: :class:`~.FieldType` enum
        fields: list of Field instances
    """

    def __init__(self, parent, key=('object', 'angle'), flds=[0.]):
        self.optical_spec = parent
        self.key = 'field', key[0], key[1]
        self.set_from_list(flds)

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['optical_spec']
        return attrs

    def sync_to_restore(self, optical_spec):
        if hasattr(self, 'field_type'):
            self.key = model_enums.get_fld_key_for_type(self.field_type)
            del self.field_type
        self.optical_spec = optical_spec

    def __str__(self):
        return "key={}, max field={}".format(self.key, self.max_field()[0])

    def set_from_list(self, flds):
        self.fields = [Field() for f in range(len(flds))]
        for i, f in enumerate(self.fields):
            f.y = flds[i]
        self.value, _ = self.max_field()

    def set_from_specsheet(self, ss):
        for k, v in ss.etendue_inputs['field'].items():
            if len(v) > 0:
                obj_img_key = k
                for k1, v1 in v.items():
                    value_key = k1
                    break

        self.key = 'field', obj_img_key, value_key
        flds = [0, ss.etendue_inputs['field'][obj_img_key][value_key]]
        self.set_from_list(flds)

    def get_input_for_specsheet(self):
        return self.key, self.max_field()[0]

    def update_model(self):
        for f in self.fields:
            f.update()

        # recalculate max_field and relabel fields.
        #  relabeling really assumes the fields are radial, specifically,
        #  y axis only
        max_field, fi = self.max_field()
        self.value = max_field
        field_norm = 1.0 if max_field == 0 else 1.0/max_field
        self.index_labels = []
        for i, f in enumerate(self.fields):
            if f.x != 0.0:
                fldx = '{:5.2f}x'.format(field_norm*f.x)
            else:
                fldx = ''
            if f.y != 0.0:
                fldy = '{:5.2f}y'.format(field_norm*f.y)
            else:
                fldy = ''
            self.index_labels.append(fldx + fldy)
        self.index_labels[0] = 'axis'
        if len(self.index_labels) > 1:
            self.index_labels[-1] = 'edge'
        return self

    def get_field_type(self):
        return model_enums.get_fld_type_for_key(self.key).value

    def mutate_field_type(self, new_field_type):
        osp = self.optical_spec
        fld_key = model_enums.get_fld_key_for_type(new_field_type)
        field, obj_img_key, value_key = fld_key
        if self.optical_spec is not None:
            if osp.parax_data is not None:
                fod = self.optical_spec.parax_data.fod
                if obj_img_key == 'object':
                    if value_key == 'height':
                        self.value = osp.parax_data.pr_ray[0][mc.ht]
                    elif value_key == 'angle':
                        self.value = fod.obj_ang
                elif obj_img_key == 'image':
                    if value_key == 'height':
                        self.value = fod.img_ht
        self.key = fld_key

    def max_field(self):
        """ calculates the maximum field of view

        Returns:
            magnitude of maximum field, maximum Field instance
        """
        max_fld = None
        max_fld_sqrd = 0.0
        for i, f in enumerate(self.fields):
            fld_sqrd = f.x*f.x + f.y*f.y
            if fld_sqrd > max_fld_sqrd:
                max_fld_sqrd = fld_sqrd
                max_fld = i
        return math.sqrt(max_fld_sqrd), max_fld


class Field:
    """ a single field point

    Attributes:
        x: x field component in absolute units
        y: y field component in absolute units
        vux: +x vignetting factor
        vuy: +y vignetting factor
        vlx: -x vignetting factor
        vly: -y vignetting factor
        wt: field weight
        aim_pt: x, y chief ray coords on the paraxial entrance pupil plane
        chief_ray: ray package for the ray from the field point throught the
                   center of the aperture stop, traced in the central
                   wavelength
        ref_sphere: a tuple containing (image_pt, cr_exp_pt, cr_exp_dist,
                    ref_dir, ref_sphere_radius)

    """

    def __init__(self, x=0., y=0., wt=1.):
        self.x = x
        self.y = y
        self.vux = 0.0
        self.vuy = 0.0
        self.vlx = 0.0
        self.vly = 0.0
        self.wt = wt
        self.aim_pt = None
        self.chief_ray = None
        self.ref_sphere = None

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['chief_ray']
        del attrs['ref_sphere']
        del attrs['pupil_rays']
        return attrs

    def __str__(self):
        return "{}, {}".format(self.x, self.y)

    def __repr__(self):
        return "Field(x={}, y={}, wt={})".format(self.x, self.y, self.wt)

    def update(self):
        self.chief_ray = None
        self.ref_sphere = None

    def apply_vignetting(self, pupil):
        vig_pupil = pupil[:]
        if pupil[0] < 0.0:
            if self.vlx != 0.0:
                vig_pupil[0] *= (1.0 - self.vlx)
        else:
            if self.vux != 0.0:
                vig_pupil[0] *= (1.0 - self.vux)
        if pupil[1] < 0.0:
            if self.vly != 0.0:
                vig_pupil[1] *= (1.0 - self.vly)
        else:
            if self.vuy != 0.0:
                vig_pupil[1] *= (1.0 - self.vuy)
        return vig_pupil


class FocusRange:
    """ Focus range specification

    Attributes:
        focus_shift: focus shift (z displacement) from nominal image interface
        defocus_range: +/- half the total focal range, from the focus_shift
                       position
    """

    def __init__(self, focus_shift=0.0, defocus_range=0.0):
        self.focus_shift = focus_shift
        self.defocus_range = defocus_range

    def __repr__(self):
        return ("FocusRange(focus_shift={}, defocus_range={})"
                .format(self.focus_shift, self.defocus_range))

    def set_from_specsheet(self, ss):
        pass

    def update(self):
        pass

    def get_focus(self, fr=0.0):
        """ return focus position for input focus range parameter

        Args:
            fr (float): focus range parameter, -1.0 to 1.0

        Returns:
            focus position for input focus range parameter
        """
        return self.focus_shift + fr*self.defocus_range



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" Module for diffractive/holographic optical elements

    The :class:`~.DiffractiveElement` and :class:`~.HolographicElement`
    implementations are patterned after Wang, et al, `Ray tracing and wave
    aberration calculation for diffractive optical elements
    <https://doi.org/10.1117/1.600780>`_

.. Created on Fri Jul  5 11:27:13 2019

.. codeauthor: Michael J. Hayford
"""


from math import sqrt
import numpy as np
import importlib

from rayoptics.util.misc_math import normalize


def radial_phase_fct(pt, coefficients):
    """ evaluate the phase and slopes at **pt** """
    x, y, z = pt
    r_sqr = x*x + y*y
    dW = 0
    dWdX = 0
    dWdY = 0
    for i, c in enumerate(coefficients):
        dW += c*r_sqr**(i+1)
        r_exp = r_sqr**(i)
        factor = 2*(i+1)
        dWdX += factor*c*x*r_exp
        dWdY += factor*c*y*r_exp
    return dW, dWdX, dWdY


class DiffractiveElement:
    def __init__(self, label='', coefficients=None, ref_wl=550., order=1,
                 phase_fct=None):
        self.label = label
        if coefficients is None:
            self.coefficients = []
        else:
            self.coefficients = coefficients
        self.ref_wl = ref_wl
        self.order = order
        self.phase_fct = phase_fct

    def __repr__(self):
        return (type(self).__name__ + '(label=' + repr(self.label) +
                ', coefficients=' + repr(self.coefficients) +
                ', ref_wl=' + repr(self.ref_wl) +
                ', order=' + repr(self.order) +
                ', phase_fct=' + repr(self.phase_fct) + ')')

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['phase_fct']
        attrs['phase_fct_module'] = self.phase_fct.__module__
        attrs['phase_fct_name'] = self.phase_fct.__name__
        return attrs

    def __json_decode__(self, **attrs):
        module_name = attrs.pop('phase_fct_module')
        fct_name = attrs.pop('phase_fct_name')
        mod = importlib.import_module(module_name)
        phase_fct = getattr(mod, fct_name)
        self.__init__(phase_fct=phase_fct, **attrs)

    def list_doe(self):
        print("ref_pt: {:12.5f} {:12.5f} {:12.5f} {}"
              .format(self.ref_pt[0], self.ref_pt[1], self.ref_pt[2],
                      self.ref_virtual))

    def phase(self, pt, in_dir, srf_nrml, wl=None):
        order = self.order
        normal = normalize(srf_nrml)
        in_cosI = np.dot(in_dir, normal)
        mu = 1.0 if wl is None else wl/self.ref_wl
        dW, dWdX, dWdY = self.phase_fct(pt, self.coefficients)
#        print(wl, mu, dW, dWdX, dWdY)
        b = in_cosI + order*mu*(normal[0]*dWdX + normal[1]*dWdY)
        c = mu*(mu*(dWdX**2 + dWdY**2)/2 +
                order*(in_dir[0]*dWdX + in_dir[1]*dWdY))
        Q = -b + sqrt(b*b - 2*c)
#        print("{:6.3f} {:12.5f} {:12.5f} {:12.5f} {:12.5f} {:12.5f} {:12.5f}"
#              .format(mu, dW, dWdX, dWdY, b, c, Q))
        out_dir = in_dir + order*mu*(np.array([dWdX, dWdY, 0])) + Q*normal
        dW *= mu
        return out_dir, dW


class HolographicElement:
    def __init__(self, label=''):
        self.label = label
        self.ref_pt = np.array([0., 0., -1e10])
        self.ref_virtual = False
        self.obj_pt = np.array([0., 0., -1e10])
        self.obj_virtual = False
        self.ref_wl = 550.0

    def list_hoe(self):
        print("ref_pt: {:12.5f} {:12.5f} {:12.5f} {}"
              .format(self.ref_pt[0], self.ref_pt[1], self.ref_pt[2],
                      self.ref_virtual))
        print("obj_pt: {:12.5f} {:12.5f} {:12.5f} {}"
              .format(self.obj_pt[0], self.obj_pt[1], self.obj_pt[2],
                      self.obj_virtual))

    def phase(self, pt, in_dir, srf_nrml, wl=None):
        normal = normalize(srf_nrml)
        ref_dir = normalize(pt - self.ref_pt)
        if self.ref_virtual:
            ref_dir = -ref_dir
        ref_cosI = np.dot(ref_dir, normal)
        obj_dir = normalize(pt - self.obj_pt)
        if self.obj_virtual:
            obj_dir = -obj_dir
        obj_cosI = np.dot(obj_dir, normal)
        in_cosI = np.dot(in_dir, normal)
        mu = 1.0 if wl is None else wl/self.ref_wl
        b = in_cosI + mu*(obj_cosI - ref_cosI)
        refp_cosI = np.dot(ref_dir, in_dir)
        objp_cosI = np.dot(obj_dir, in_dir)
        ro_cosI = np.dot(ref_dir, obj_dir)
        c = mu*(mu*(1.0 - ro_cosI) + (objp_cosI - refp_cosI))
        Q = -b + sqrt(b*b - 2*c)
        out_dir = in_dir + mu*(obj_dir - ref_dir) + Q*normal
        dW = 0.
        return out_dir, dW



""" package implementing useful rayoptics graphics using matplotlib

    The ``rayoptics.mpl`` subpackage provides useful basic optical graphics
    using the matplotlib plotting package. Particular features include:

        - 2D lens layout
        - ray aberration and wavefront pupil/field plots
        - |ybar| and |nubar| paraxial ray diagrams
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 18 10:59:29 2017

@author: Mike
"""


import unittest
import numpy as np
import numpy.testing as npt
from math import log10, pow
from rayoptics.optical.sequential import gen_sequence
from rayoptics.optical.raytrace import trace_raw
from rayoptics.optical.traceerror import TraceError
from rayoptics.util.misc_math import normalize

import ag_dblgauss_s as dblg
import marginal_ray as f1r2


class RayTraceTestCase(unittest.TestCase):
    def setUp(self):
        # lump defocus into back focal distance
        dblg.ag_dblgauss[-2][1] += dblg.ag_dblgauss[-1][1]
        dblg.ag_dblgauss[-1][1] = 0.

        self.p0 = np.array([0., 0., 0.])
        self.p1 = np.array([0., 0., dblg.ag_dblgauss[0][1]])
        self.epd = np.array([25., 0., 0.])
        self.d0 = normalize((self.p1 + self.epd) - self.p0)

    def test_dbgauss_axial_ray(self):
        wv = 587.6
        seq = gen_sequence(dblg.ag_dblgauss, wvl=wv, radius_mode=False)
        try:
            ray, op_delta, wvl = trace_raw(seq, self.p0, self.d0, wv)
        except TraceError as ray_error:
            ray = ray_error.ray

        def rel_err(v, def_err):
            if v == 0.:
                return def_err
            else:
                vmag = round(log10(abs(v)), 1)
                if vmag < 0.:
                    return def_err/pow(10, vmag)
                else:
                    return def_err

        def rel_err_vec(v, def_err):
            vflt = list(abs(x) for x in v if x != 0)
            if len(vflt) == 0:
                return def_err
            else:
                return rel_err(min(vflt), def_err)

        for i, rtup in enumerate(zip(ray, f1r2.rayf1r2)):
            xyztol = rel_err_vec(rtup[1][0], 1e-6)
#            print(i, rtup[0][0], rtup[1][0], xyztol)
            npt.assert_allclose(rtup[0][0], rtup[1][0], rtol=xyztol)
            # compute direction tangents from direction cosines
            tantol = rel_err_vec(rtup[1][1], 1e-6)
            npt.assert_allclose([rtup[0][1][0]/rtup[0][1][2],
                                 rtup[0][1][1]/rtup[0][1][2]], rtup[1][1],
                                rtol=tantol, atol=1e-10)

            # don't compare path length in object or image space
            if i > 1 and i < 12:
                dsttol = rel_err(rtup[1][2], 1e-6)
                npt.assert_allclose(rtup[0][2], rtup[1][2], rtol=dsttol)


if __name__ == '__main__':
    unittest.main(verbosity=3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Support for fully featured QT windows for plotting/matplotlib

.. Created on Wed Nov  7 15:04:19 2018

.. codeauthor: Michael J. Hayford
"""
from pathlib import Path

from PyQt5.QtCore import Qt as qt
from PyQt5.QtCore import QSize
from PyQt5 import QtGui
from PyQt5.QtWidgets import (QHBoxLayout, QVBoxLayout, QWidget, QLineEdit,
                             QRadioButton, QGroupBox, QSizePolicy, QCheckBox,
                             QListWidget, QListWidgetItem, QToolBar)

from matplotlib.backends.backend_qt5agg \
     import (NavigationToolbar2QT as NavigationToolbar)
from matplotlib.backends.backend_qt5agg \
     import FigureCanvasQTAgg as FigureCanvas

from rayoptics.gui.appmanager import ModelInfo
from rayoptics.mpl.axisarrayfigure import Fit


class PlotCanvas(FigureCanvas):

    def __init__(self, parent, figure):
        FigureCanvas.__init__(self, figure)
        self.setParent(parent)

        # Next 2 lines are needed so that key press events are correctly
        #  passed with mouse events
        # https://github.com/matplotlib/matplotlib/issues/707/
        self.setFocusPolicy(qt.ClickFocus)
        self.setFocus()

        FigureCanvas.setSizePolicy(self,
                                   QSizePolicy.Expanding,
                                   QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)

        self.figure.plot()


def update_figure_view(plotFigure):
    plotFigure.refresh()


class CommandItem(QListWidgetItem):
    def __init__(self, parent, txt, cntxt):
        super().__init__(parent)
        self.setData(qt.DisplayRole, txt)
        self.setData(qt.EditRole, cntxt)

    def data(self, role):
        if role == qt.DisplayRole:
            return self.txt
        elif role == qt.EditRole:
            return self.cntxt
        else:
            return None

    def setData(self, role, value):
        if role == qt.DisplayRole:
            self.txt = value
            return True
        elif role == qt.EditRole:
            self.cntxt = value
            return True
        else:
            return False


def create_command_panel(fig, commands):
    command_panel = QListWidget()

    for c in commands:
        cmd_txt, cntxt = c
        cntxt[2]['figure'] = fig
        btn = CommandItem(command_panel, cmd_txt, cntxt)

    command_panel.itemClicked.connect(on_command_clicked)
    width = command_panel.size()
    hint = command_panel.sizeHint()
    frame_width = command_panel.frameWidth() + 2
    column_width = command_panel.sizeHintForColumn(0) + 2*frame_width
    command_panel.setMinimumWidth(column_width)
    command_panel.setMaximumWidth(column_width)
    command_panel.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

    return command_panel


def on_command_clicked(item):
    cntxt = item.data(qt.EditRole)
    fct, args, kwargs = cntxt
    fct(*args, **kwargs)


def create_plot_view(app, fig, title, view_width, view_ht, commands=None,
                     add_panel_fcts=None, add_nav_toolbar=False):
    """ create a window hosting a (mpl) figure """

    def create_light_or_dark_callback(fig):
        def l_or_d(is_dark):
            fig.sync_light_or_dark(is_dark)
        return l_or_d

    # construct the top level widget
    widget = QWidget()

    top_layout = QHBoxLayout(widget)
    # set the layout on the widget
    widget.setLayout(top_layout)

    if commands:
        command_panel = create_command_panel(fig, commands)
        top_layout.addWidget(command_panel)

    # construct the top level layout
    plot_layout = QVBoxLayout()
    top_layout.addLayout(plot_layout)

    pc = PlotCanvas(app, fig)
    if add_panel_fcts is not None:
        panel_layout = QHBoxLayout()
        plot_layout.addLayout(panel_layout)
        for p in add_panel_fcts:
            panel = p(app, pc)
            panel_layout.addWidget(panel)
        panel_layout.addStretch(50)

    mi = ModelInfo(app.app_manager.model, update_figure_view, (fig,))
    sub_window = app.add_subwindow(widget, mi)
    sub_window.sync_light_or_dark = create_light_or_dark_callback(fig)
    sub_window.setWindowTitle(title)
    orig_x, orig_y = app.initial_window_offset()
    sub_window.setGeometry(orig_x, orig_y, view_width, view_ht)

    plot_layout.addWidget(pc)

    if add_nav_toolbar:
        plot_layout.addWidget(NavigationToolbar(pc, sub_window))

    sub_window.show()


def create_plot_scale_panel(app, pc):
    groupBox = QGroupBox("Plot Scale", app)

    user_scale_wdgt = QLineEdit()
    user_scale_wdgt.setReadOnly(True)
    pf = pc.figure
    cntxt = pf, user_scale_wdgt
    user_scale_wdgt.editingFinished.connect(lambda:
                                            on_plot_scale_changed(cntxt))
    fit_all_btn = QRadioButton("Fit All")
    fit_all_btn.setChecked(pf.scale_type == Fit.All)
    fit_all_btn.toggled.connect(lambda:
                                on_plot_scale_toggled(cntxt, Fit.All))
    fit_all_same_btn = QRadioButton("Fit All Same")
    fit_all_same_btn.setChecked(pf.scale_type == Fit.All_Same)
    fit_all_same_btn.toggled.connect(lambda: on_plot_scale_toggled(
                                             cntxt, Fit.All_Same))
    user_scale_btn = QRadioButton("User Scale")
    user_scale_btn.setChecked(pf.scale_type == Fit.User_Scale)
    user_scale_btn.toggled.connect(lambda: on_plot_scale_toggled(
                                           cntxt, Fit.User_Scale))
    box = QHBoxLayout()
    box.addWidget(fit_all_btn)
    box.addWidget(fit_all_same_btn)
    box.addWidget(user_scale_btn)
    box.addWidget(user_scale_wdgt)

    groupBox.setLayout(box)

    return groupBox


def on_plot_scale_toggled(cntxt, scale_type):
    plotFigure, scale_wdgt = cntxt
    plotFigure.scale_type = scale_type
    if scale_type == Fit.User_Scale:
        scale_wdgt.setReadOnly(False)
        scale_wdgt.setText('{:7.4f}'.format(plotFigure.user_scale_value))
    else:
        scale_wdgt.setReadOnly(True)

    plotFigure.plot()


def on_plot_scale_changed(cntxt):
    plotFigure, scale_wdgt = cntxt
    eval_str = scale_wdgt.text()
    try:
        val = eval(eval_str)
        plotFigure.user_scale_value = val
        scale_wdgt.setText('{:7.4f}'.format(val))
    except IndexError:
        return ''

    plotFigure.plot()


def get_icon(fig, icon_filepath, icon_size=48):
    pm = QtGui.QPixmap(str(icon_filepath)).scaled(icon_size, icon_size)
    # pm = QtGui.QPixmap(str(icon_filepath))
    if hasattr(pm, 'setDevicePixelRatio'):
        pm.setDevicePixelRatio(fig.canvas._dpi_ratio)
    return QtGui.QIcon(pm)


def create_2d_figure_toolbar(app, pc):
    """ zoom, fit and pan commands for figures
    Pan
    Zoom Box
    Fit
    Zoom In, Out
    1:1
    """
    icon_size = 24
    tb = QToolBar()
    tb.setIconSize(QSize(icon_size, icon_size))
    tb.setStyleSheet("QToolBar{spacing:0px;}")

    toolitems = (
        ('Pan', 'Pan axes with mouse', 'pan', 'register_pan'),
        ('Zoom', 'Zoom to rectangle', 'zoom', 'register_zoom_box'),
        ('Fit', 'Fit to view', 'fit', 'fit'),
        ('Zoom In', 'Zoom in', 'zoom_in', 'zoom_in'),
        ('Zoom Out', 'Zoom out', 'zoom_out', 'zoom_out'),
      )

    pth = Path(__file__).resolve().parent
    image_dir = Path(pth / 'images')

    fig = pc.figure
    actions = {}
    for text, tooltip_text, image_file, callback in toolitems:
        if text is None:
            tb.addSeparator()
        else:
            def create_action_and_on_finished(fig, callback):
                def make_cb_fct():
                    def on_finished():
                        # unpress button when action is finished
                        actions[callback].setChecked(False)
                    getattr(fig, callback)(on_finished)
                return make_cb_fct

            image_path = image_dir / (image_file + '.png')
            icon = get_icon(fig, image_path, icon_size=icon_size)

            if callback in ['register_zoom_box', 'register_pan']:
                # action requires mouse input handling, unpress when finished
                a = tb.addAction(icon, text,
                                 create_action_and_on_finished(fig, callback))
                actions[callback] = a
                a.setCheckable(True)
            else:
                # immediate action button
                a = tb.addAction(icon, text, getattr(fig, callback))
                actions[callback] = a

            if tooltip_text is not None:
                a.setToolTip(tooltip_text)

    def attr_check(fig, attr, state):
        checked = state == qt.Checked
        setattr(fig, attr, checked)
        fig.refresh()

    # add checkbox for unit aspect ration display
    aratio_checkBox = QCheckBox("1:1")
    aratio_checkBox.setChecked(fig.is_unit_aspect_ratio)
    aratio_checkBox.stateChanged.connect(
        lambda checked: attr_check(fig, 'is_unit_aspect_ratio', checked))
    tb.addWidget(aratio_checkBox)

    return tb


def create_draw_rays_groupbox(app, pc):
    tb = QToolBar()
    fig = pc.figure

    def attr_check(fig, attr, state):
        checked = state == qt.Checked
#        cur_value = getattr(fig, attr, None)
        setattr(fig, attr, checked)
        fig.refresh()

    parax_checkBox = QCheckBox("&paraxial rays")
    parax_checkBox.setChecked(fig.do_paraxial_layout)
    parax_checkBox.stateChanged.connect(
        lambda checked: attr_check(fig, 'do_paraxial_layout', checked))
    edge_checkBox = QCheckBox("&edge rays")
    edge_checkBox.setChecked(fig.do_draw_rays)
    edge_checkBox.stateChanged.connect(
        lambda checked: attr_check(fig, 'do_draw_rays', checked))

    tb.addWidget(parax_checkBox)
    tb.addWidget(edge_checkBox)

    return tb


def create_diagram_controls_groupbox(app, pc):
    tb = QToolBar()

    def attr_check(fig, attr, state):
        checked = state == qt.Checked
        # cur_value = getattr(fig, attr, None)
        setattr(fig, attr, checked)
        # print('attr_check: {}={}'.format(attr, checked))
        fig.refresh()

    barrel_value_wdgt = QLineEdit()
    barrel_value_wdgt.setReadOnly(True)
    fig = pc.figure
    cntxt = fig, barrel_value_wdgt
    barrel_value_wdgt.editingFinished.connect(lambda:
                                              on_barrel_constraint_changed(
                                                      cntxt))

    slide_checkBox = QCheckBox("&slide")
    slide_checkBox.setChecked(fig.enable_slide)
    slide_checkBox.stateChanged.connect(
        lambda checked: attr_check(fig, 'enable_slide', checked))
    barrel_checkBox = QCheckBox("&barrel constraint")
    barrel_checkBox.setChecked(fig.diagram.do_barrel_constraint)
    barrel_checkBox.stateChanged.connect(
        lambda checked: on_barrel_constraint_toggled(cntxt, checked))

    diagram = fig.diagram
    bending_btn = QRadioButton("Bend")
    bending_btn.setChecked(diagram.bend_or_gap == 'bend')
    bending_btn.toggled.connect(lambda:
                                on_bend_or_gap_toggled(diagram, 'bend'))
    gap_btn = QRadioButton("Gap")
    gap_btn.setChecked(diagram.bend_or_gap == 'gap')
    gap_btn.toggled.connect(lambda: on_bend_or_gap_toggled(diagram, 'gap'))

    tb.addWidget(slide_checkBox)
    tb.addWidget(barrel_checkBox)
    tb.addWidget(barrel_value_wdgt)
    tb.addWidget(bending_btn)
    tb.addWidget(gap_btn)

    return tb


def on_barrel_constraint_toggled(cntxt, state):
    fig, barrel_wdgt = cntxt
    diagram = fig.diagram
    checked = state == qt.Checked
    if checked:
        diagram.do_barrel_constraint = True
        barrel_wdgt.setReadOnly(False)
        barrel_wdgt.setText('{:7.4f}'.format(diagram.barrel_constraint_radius))
    else:
        diagram.do_barrel_constraint = False
        barrel_wdgt.setReadOnly(True)

    fig.refresh()


def on_barrel_constraint_changed(cntxt):
    fig, barrel_wdgt = cntxt
    eval_str = barrel_wdgt.text()
    try:
        val = eval(eval_str)
        fig.diagram.barrel_constraint_radius = val
        barrel_wdgt.setText('{:7.4f}'.format(val))
    except IndexError:
        return ''

    fig.refresh()


def on_bend_or_gap_toggled(diagram, radio_btn_id):
    diagram.bend_or_gap = radio_btn_id



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" Interactive layout figure with paraxial editing

.. Created on Thu Mar 14 10:20:33 2019

.. codeauthor: Michael J. Hayford
"""

import numpy as np

from rayoptics.gui.util import bbox_from_poly

from rayoptics.mpl.interactivefigure import InteractiveFigure

from rayoptics.gui.layout import LensLayout


class InteractiveLayout(InteractiveFigure):
    """ Editable version of optical system layout, aka Live Layout

    Attributes:
        opt_model: parent optical model
        refresh_gui: function to be called on refresh_gui event
        offset_factor: how much to draw rays before first surface
        do_draw_rays: if True, draw edge rays
        do_paraxial_layout: if True, draw editable paraxial axial and chief ray
    """

    def __init__(self, opt_model, refresh_gui,
                 offset_factor=0.05,
                 do_draw_rays=False,
                 do_paraxial_layout=True,
                 **kwargs):
        self.refresh_gui = refresh_gui
        self.layout = LensLayout(opt_model)
        self.linewidth = 0.5
        self.do_draw_rays = do_draw_rays
        self.do_paraxial_layout = do_paraxial_layout
        self.offset_factor = offset_factor
        if 'do_scale_bounds' not in kwargs:
            kwargs['do_scale_bounds'] = True

        super().__init__(**kwargs)

    def update_data(self):
        self.artists = []
        concat_bbox = []
        layout = self.layout

        self.ele_shapes = layout.create_element_model(self)
        self.ele_bbox = self.update_patches(self.ele_shapes)
        concat_bbox.append(self.ele_bbox)

        if self.do_draw_rays:
            sl_so = layout.system_length(self.ele_bbox,
                                         offset_factor=self.offset_factor)
            system_length, start_offset = sl_so
            self.ray_shapes = layout.create_ray_model(self, start_offset)
            self.ray_bbox = self.update_patches(self.ray_shapes)
            concat_bbox.append(self.ray_bbox)

        if self.do_paraxial_layout:
            self.parax_shapes = layout.create_paraxial_layout(self)
            self.parax_bbox = self.update_patches(self.parax_shapes)
            concat_bbox.append(self.parax_bbox)

        sys_bbox = np.concatenate(concat_bbox)
        self.sys_bbox = bbox_from_poly(sys_bbox)

        return self

    def action_complete(self):
        super().action_complete()
        self.do_action = self.do_shape_action

    def fit_axis_limits(self):
        return self.sys_bbox



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
""" manage light and dark interface color schemes

.. Created on Mon Feb  3 22:03:16 2020

.. codeauthor: Michael J. Hayford
"""
from pathlib import Path

from matplotlib import style
from matplotlib.figure import Figure

from rayoptics.util import colors


def apply_style(is_dark):
    """ Assign a light or dark style to mpl plots. """
    pth = Path(__file__).resolve().parent
    styles_dir = Path(pth / 'styles')
    if is_dark:
        style_path = styles_dir / 'Solarize_Dark.mplstyle'
        style.use(str(style_path))
    else:
        style_path = styles_dir / 'Solarize_Light.mplstyle'
        style.use(str(style_path))


class StyledFigure(Figure):
    """ Provide a standard implementation for mpl styles. """

    def __init__(self, **kwargs):
        is_dark = kwargs.pop('is_dark', False)
        self._rgb = {**colors.accent_colors(is_dark),
                     **colors.foreground_background(is_dark)}

        super().__init__(**kwargs)

        self.sync_light_or_dark(is_dark, do_refresh=False)

    def sync_light_or_dark(self, is_dark, do_refresh=True):
        self._rgb = {**colors.accent_colors(is_dark),
                     **colors.foreground_background(is_dark)}
        apply_style(is_dark)
        self.set_facecolor(self._rgb['background'])
        if hasattr(self, 'ax'):
            axes = self.ax
            axes.set_facecolor(self._rgb['background1'])
            axes.tick_params(
                colors=self._rgb['foreground'],
                grid_color=self._rgb['background']
                )
        if do_refresh:
            self.refresh()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" Definitions for user interface actions

    `Action` classes are those that contain a `dict` called ``actions``. The
    ``actions`` dict contains functions that are executed for UI actions 
    corresponding to the keyword.

    Action keywords include:
        'press'
        'drag'
        'release'

    The argument lists for the called functions include ``fig`` and ``event``.
    Duck typing is relied on loosen the binding with a particular GUI
    implementation - or at least that's the theory.

    ``fig`` must support a function refresh_gui()

.. Created on Wed Apr 24 14:26:14 2019

.. codeauthor: Michael J. Hayford
"""

import math


class Action():
    """ Action built on a set/get function pair """
    def __init__(self, getf, setf):
        self.getf = getf
        self.setf = setf
        self.cur_value = None
        self.new_value = None
        self.actions = {}

        def on_select(fig, event):
            self.cur_value = getf()
            return self.cur_value
        self.actions['press'] = on_select

        def on_edit(fig, event, new_value):
            setf(self.cur_value, new_value)
            fig.refresh_gui()
        self.actions['drag'] = on_edit

        def on_release(fig, event):
            self.new_value = getf()
            fig.refresh_gui()
        self.actions['release'] = on_release


class AttrAction():
    """ Action built on an object/attribute pair """
    def __init__(self, obj, attr):
        self.object = obj
        self.attr = attr
        self.cur_value = getattr(self.object, self.attr, None)
        self.new_value = None
        self.actions = {}

        def on_select(fig, event):
            self.cur_value = getattr(self.object, self.attr, None)
#            print('AttrAction.on_select:', self.attr, self.cur_value)
            return self.cur_value
        self.actions['press'] = on_select

        def on_edit(fig, event, delta_value):
            setattr(self.object, self.attr, self.cur_value+delta_value)
#            print('AttrAction.on_edit:', self.attr, self.cur_value+delta_value)
            fig.refresh_gui()
        self.actions['drag'] = on_edit

        def on_release(fig, event):
            self.new_value = getattr(self.object, self.attr, None)
            fig.refresh_gui()
        self.actions['release'] = on_release


class SagAction():
    """ Action to set the profile_cv via surface sag (related to x,y input) """
    def __init__(self, surf):
        self.surf = surf
        self.cur_value = None
        self.new_value = None
        self.actions = {}

        def on_select(fig, event):
            self.cur_value = self.surf.z_sag((event.x, event.ydata))
#            print('SagAction.on_select:', self.cur_value)
            return self.cur_value
        self.actions['press'] = on_select

        def on_edit(fig, event, value):
            self.surf.set_z_sag(value)
#            cv = self.surf.calc_cv_from_zsag(value)
#            print('SagAction.on_edit (x, y, cv):', value, cv)
            fig.refresh_gui()
        self.actions['drag'] = on_edit

        def on_release(fig, event):
            self.new_value = self.surf.z_sag((event.x, event.ydata))
            fig.refresh_gui()
        self.actions['release'] = on_release


class BendAction():
    """ Action to bend a lens element, using x component of user input """
    def __init__(self, e):
        self.ele = e
        self.select_pt_lcl = None
        self.cv_orig = None
        self.cv_new = None
        self.actions = {}

        def sag(cv, y):
            if cv != 0.0:
                r = 1/cv
                adj = math.sqrt(r*r - y*y)
                return r*(1 - abs(adj/r))
            else:
                return 0

        def on_select(fig, event):
            self.select_pt_lcl = event.lcl_pt
            self.cv_orig = self.ele.reference_interface().profile_cv
            self.xsag_orig = sag(self.cv_orig, event.lcl_pt[1])
#            print('on_select: {:.3f} {:.3f} {:.5f}'
#                  .format(event.lcl_pt[0], self.xsag_orig, self.cv_orig))
            return self.select_pt_lcl
        self.actions['press'] = on_select

        def on_edit(fig, event, lcl_pt):
            cv1 = self.ele.s1.profile_cv
            cv2 = self.ele.s2.profile_cv
            x, y = lcl_pt
            xsag_new = self.xsag_orig + (x - self.select_pt_lcl[0])
            new_pt = xsag_new, lcl_pt[1]
            cv1_new = self.ele.reference_interface().calc_cv_from_zsag(new_pt)
#            print('on_edit: {:.3f} {:.3f} {:.5f}'.format(x, xsag_new, cv1_new))
            delta_cv = cv1_new - cv1
            cv2_new = cv2 + delta_cv
            self.ele.s1.profile_cv = cv1_new
            self.ele.s2.profile_cv = cv2_new
            fig.refresh_gui()
        self.actions['drag'] = on_edit

        def on_release(fig, event):
            self.cv_new = self.ele.reference_interface().profile_cv
            xsag = sag(self.cv_new, event.lcl_pt[1])
#            print('on_release: {:.3f} {:.3f} {:.5f}'
#                  .format(event.lcl_pt[0], xsag, self.cv_new))
            fig.refresh_gui()
        self.actions['release'] = on_release



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" mpl implementations of common optical analyses

.. codeauthor: Michael J. Hayford
"""

import abc

import numpy as np

from rayoptics.mpl.axisarrayfigure import Fit
from rayoptics.mpl.styledfigure import StyledFigure

from rayoptics.optical.opticalspec import Field
from rayoptics.optical.trace import trace_astigmatism
from rayoptics.optical.thirdorder import compute_third_order


class FieldCurveFigure(StyledFigure):
    """ Plot of astigmatism curves """

    def __init__(self, opt_model,
                 eval_fct=trace_astigmatism,
                 **kwargs):
        self.opt_model = opt_model
        self.scale_type = Fit.All
        self.eval_fct = eval_fct

        super().__init__(**kwargs)

        self.update_data()

    def refresh(self):
        self.update_data()
        self.plot()

    def update_data(self):
        self.s_data = []
        self.t_data = []
        self.field_data = []

        osp = self.opt_model.optical_spec
        _, wvl, foc = osp.lookup_fld_wvl_focus(0)
        fld = Field()
        max_field = osp.field_of_view.max_field()[0]
        for f in np.linspace(0., max_field, num=11):
            fld.y = f
            s_foc, t_foc = self.eval_fct(self.opt_model, fld, wvl, foc)
            self.s_data.append(s_foc)
            self.t_data.append(t_foc)
            self.field_data.append(f)
        return self

    def plot(self):
        self.clf()
        self.ax = self.add_subplot(1, 1, 1)

        self.ax.set_title("Astigmatic Field Plot", pad=10.0, fontsize=18)

        self.ax.plot(self.s_data, self.field_data,
                     label='sagittal',
                     c=self._rgb['violet'])
        self.ax.plot(self.t_data, self.field_data,
                     label='tangential',
                     c=self._rgb['magenta'])
        self.ax.set_xlabel('focus')
        self.ax.set_ylabel('field height')

        self.ax.legend()

        self.canvas.draw()

        return self


class ThirdOrderBarChart(StyledFigure):
    def __init__(self, opt_model,
                 **kwargs):
        super().__init__(**kwargs)
        self.opt_model = opt_model
        self.scale_type = Fit.All

        self.update_data()

    def refresh(self):
        self.update_data()
        self.plot()

    def update_data(self):
        self.to_pkg = compute_third_order(self.opt_model)

    def plot(self):
        self.clf()
        self.ax = self.add_subplot(1, 1, 1)

        self.ax.set_xlabel('Surface')
        self.ax.set_ylabel('third order aberration')
        self.ax.set_title('Surface by surface third order aberrations')
        self.to_pkg.plot.bar(ax=self.ax, rot=0)
        self.ax.grid(True)
        self.tight_layout()

        self.canvas.draw()

        return self


# experimental - something usable from qt and jupyter
class AnalysisPlot(abc.ABC):
    """ abstract api for matplotlib axes customized for specific analyses """

    def __init__(self, opt_model):
        self.opt_model = opt_model

    def refresh(self):
        """ called by the app manager to refresh the plot """
        self.update_data()
        self.plot()

    @abc.abstractmethod
    def update_data(self):
        """ function to update the backend data needed for the plot """

    @abc.abstractmethod
    def plot(self):
        """ function that executes the plotting commands """


class AstigmatismCurvePlot(AnalysisPlot):
    def __init__(self, opt_model, eval_fct=trace_astigmatism, **kwargs):
        super().__init__(opt_model)
        self.scale_type = Fit.All
        self.eval_fct = eval_fct

        self.update_data()

    def update_data(self):
        self.s_data = []
        self.t_data = []
        self.field_data = []

        osp = self.opt_model.optical_spec
        _, wvl, foc = osp.lookup_fld_wvl_focus(0)
        fld = Field()
        max_field = osp.field_of_view.max_field()[0]
        for f in np.linspace(0., max_field, num=11):
            fld.y = f
            s_foc, t_foc = self.eval_fct(self.opt_model, fld, wvl, foc)
            self.s_data.append(s_foc)
            self.t_data.append(t_foc)
            self.field_data.append(f)

    def plot(self):
        self.ax.cla()

        self.ax.set_title("Astigmatic Field Plot", pad=10.0, fontsize=18)

        self.ax.plot(self.s_data, self.field_data, label='sagittal')
        self.ax.plot(self.t_data, self.field_data, label='tangential')
        self.ax.set_xlabel('focus')
        self.ax.set_ylabel('field height')

        self.ax.legend()

#        fig.canvas.draw()



""" Package encompassing platform agnostic gui commands and updating

    The ``rayoptics.gui`` subpackage provides a thin app manager framework
    :mod:`~.appmanager` and functions implementing basic commands.

    The :mod:`~.layout` module provides support for interactive editing of
    optical layouts. This includes support for selecting and drawing different
    graphical parts of the optical model.
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" optical model enums

.. Created on Tue Dec  4 11:32:57 2018

.. codeauthor: Michael J. Hayford
"""

from enum import Enum


class PupilType(Enum):
    """ enum for different aperture specifications """
    EPD = 0  #: entrance pupil diameter
    NAO = 1  #: object space numerical aperture
    FNO = 2  #: image space f/#
    NA = 3   #: image space numerical aperture


def get_ape_key_for_type(pupil_type):
    if pupil_type == PupilType.EPD:
        obj_img_key = 'object'
        value_key = 'pupil'
    elif pupil_type == PupilType.NAO:
        obj_img_key = 'object'
        value_key = 'NA'
    elif pupil_type == PupilType.FNO:
        obj_img_key = 'image'
        value_key = 'f/#'
    elif pupil_type == PupilType.NA:
        obj_img_key = 'image'
        value_key = 'NA'
    return 'aperture', obj_img_key, value_key


def get_ape_type_for_key(aperture_key):
    aperture, obj_img_key, value_key = aperture_key
    if obj_img_key == 'object':
        if value_key == 'pupil':
            pupil_type = PupilType.EPD
        elif value_key == 'NA':
            pupil_type = PupilType.NAO
    elif obj_img_key == 'image':
        if value_key is 'NA':
            pupil_type = PupilType.NA
        elif value_key == 'f/#':
            pupil_type = PupilType.FNO
    return pupil_type


class FieldType(Enum):
    """ enum for different field specifications """
    OBJ_ANG = 0  #: object space angle in degrees
    OBJ_HT = 1   #: object height
    IMG_HT = 2   #: image height
    IMG_ANG = 3   #: image space angle in degrees


def get_fld_key_for_type(field_type):
    if field_type == FieldType.OBJ_HT:
        obj_img_key = 'object'
        value_key = 'height'
    elif field_type == FieldType.OBJ_ANG:
        obj_img_key = 'object'
        value_key = 'angle'
    elif field_type == FieldType.IMG_HT:
        obj_img_key = 'image'
        value_key = 'height'
    elif field_type == FieldType.IMG_ANG:
        obj_img_key = 'image'
        value_key = 'angle'
    return 'field', obj_img_key, value_key


def get_fld_type_for_key(field_key):
    field, obj_img_key, value_key = field_key
    if obj_img_key == 'object':
        if value_key == 'height':
            field_type = FieldType.OBJ_HT
        elif value_key == 'angle':
            field_type = FieldType.OBJ_ANG
    elif obj_img_key == 'image':
        if value_key == 'height':
            field_type = FieldType.IMG_HT
    return field_type


class DimensionType(Enum):
    """ enum for different linear dimensions """
    MM = 0  #: millimeters
    CM = 1  #: centimeters
    M = 2   #: meters
    IN = 3  #: inches
    FT = 4  #: feet


class DecenterType(Enum):
    """ enum for different tilt and decenter types """
    LOCAL = 0  #: pos and orientation applied prior to surface
    REV = 1    #: pos and orientation applied following surface in reverse
    DAR = 2    #: pos and orientation applied prior to surface and then returned to initial frame
    BEND = 3   #: used for fold mirrors, orientation applied before and after surface



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
"""

.. Created on Thu Oct 10 22:02:44 2019

.. codeauthor: Michael J. Hayford
"""

from rayoptics.gui.diagram import Diagram
from rayoptics.mpl.interactivefigure import InteractiveFigure


class InteractiveDiagram(InteractiveFigure):
    """ Editable version of optical system layout, aka Live Layout

    Attributes:
        opt_model: parent optical model
        refresh_gui: function to be called on refresh_gui event
        dgm_type: diagram type, 'ht' or 'slp'
    """

    def __init__(self, opt_model, refresh_gui, dgm_type,
                 do_barrel_constraint=False, barrel_constraint=1.0,
                 enable_slide=False, bend_or_gap='bend', **kwargs):
        self.refresh_gui = refresh_gui
        self.parax_model = opt_model.parax_model
        is_dark = kwargs['is_dark'] if 'is_dark' in kwargs else False
        self.diagram = Diagram(opt_model, dgm_type,
                               do_barrel_constraint=do_barrel_constraint,
                               barrel_constraint=barrel_constraint,
                               bend_or_gap=bend_or_gap, is_dark=is_dark)
        self.setup_dgm_type(dgm_type)
        self.enable_slide = enable_slide

        self.build = 'rebuild'

        super().__init__(**kwargs)

    def setup_dgm_type(self, dgm_type):
        if dgm_type == 'ht':
            self.x_label = r'$\overline{y}$'
            self.y_label = 'y'
            self.header = r'$y-\overline{y}$ Diagram'
        elif dgm_type == 'slp':
            self.x_label = r'$\overline{\omega}$'
            self.y_label = r'$\omega$'
            self.header = r'$\omega-\overline{\omega}$ Diagram'

    def sync_light_or_dark(self, is_dark, **kwargs):
        self.diagram.sync_light_or_dark(is_dark)
        super().sync_light_or_dark(is_dark, **kwargs)

    def update_data(self):
        self.artists = []
        self.sys_bbox = self.diagram.update_data(self)
        self.build = 'rebuild'
        return self

    def action_complete(self):
        super().action_complete()
        self.diagram.register_commands((), figure=self)

    def fit_axis_limits(self):
        return self.diagram.fit_axis_limits()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Table model supporting data content via python eval() fct

.. Created on Wed Mar 14 21:59:43 2018

.. codeauthor: Michael J. Hayford
"""
import logging

from PyQt5.QtCore import Qt, QAbstractTableModel
from PyQt5.QtCore import pyqtSignal

from rayoptics.util.misc_math import isanumber


class PyTableModel(QAbstractTableModel):
    """Table model supporting data content via python eval() fct.

    Model interface for table view of list structures.

    Attributes:
        root: object or list at the root of the eval() string
        rootEvalStr: string that is concatentated to the root name and
                    passed to the eval() function. This will accomodate
                    dynamic name changes.
        colEvalStr: string that is concatentated to the root name and
                    passed to the eval() function. There should be a
                    replacement field, i.e. {} where the row value will
                    be substituted using the str.format() function.
        rowHeaders: list of strings, length defines number of rows in the
                    table
        colHeaders: list of strings, length defines number of columns in
                    the table
        colFormats: format strings to be used to format data in each column
        is_editable: if true, items are editable
        get_num_rows: if not None, a function that returns the number of
                      rows in the table
        get_row_headers: if not None, a function that returns the row
                         headers for the table
    """

    update = pyqtSignal(object, int)

    def __init__(self, root, rootEvalStr, colEvalStr, rowHeaders,
                 colHeaders, colFormats, is_editable=False, get_num_rows=None,
                 get_row_headers=None):
        """
    """
        super().__init__()
        self.root = root
        self.rootEvalStr = rootEvalStr
        self.colEvalStr = colEvalStr
        self.rowHeaders = rowHeaders
        self.colHeaders = colHeaders
        self.colFormats = colFormats
        self.is_editable = is_editable
        self.get_num_rows = get_num_rows
        self.get_row_headers = get_row_headers

    def rowCount(self, index):
        if self.get_num_rows is not None:
            return self.get_num_rows()
        else:
            return len(self.rowHeaders)

    def columnCount(self, index):
        return len(self.colHeaders)

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return self.colHeaders[section]
            elif orientation == Qt.Vertical:
                if self.get_row_headers is not None:
                    self.rowHeaders = self.get_row_headers()
                if len(self.rowHeaders) == 0:
                    return None
                return self.rowHeaders[section]
        else:
            return None

    def flags(self, index):
        base_flag = Qt.ItemIsEnabled | Qt.ItemIsSelectable
        if self.is_editable:
            return base_flag | Qt.ItemIsEditable
        else:
            return base_flag

    def get_root_object(self):
        if len(self.rootEvalStr) == 0:
            return self.root
        else:
            root_eval_str = ('self.root' + self.rootEvalStr)
            try:
                root = eval(root_eval_str)
                return root
            except IndexError:
                return self.root

    def data(self, index, role):
        root = self.get_root_object()
        if role == Qt.DisplayRole or role == Qt.EditRole:
            r = index.row()
            c = index.column()
            eval_str = ('root' + self.colEvalStr[c]).format(r)
            try:
                val = eval(eval_str)
                valStr = self.colFormats[c].format(val)
                return valStr
            except IndexError:
                return ''
            except TypeError:
                print('Data type error: ', eval_str, val)
                return ''
        else:
            return None

    def setData(self, index, value, role):
        root = self.get_root_object()
        if role == Qt.EditRole:
            r = index.row()
            c = index.column()
            exec_str = ('root' + self.colEvalStr[c]).format(r)
            if not isanumber(value):
                value = "'" + value + "'"
            exec_str = exec_str + '=' + value
            try:
                exec(exec_str)
                self.update.emit(root, r)
                return True
            except IndexError:
                return False
            except SyntaxError:
                logging.info('Syntax error: "%s"', value)
                return False
        else:
            return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" calculate conic constants for different 2 mirror configurations

.. Created on Tue Jun 26 16:11:58 2018
.. |ParaxialModel| replace:: :class:`~.paraxialdesign.ParaxialModel`

.. codeauthor: Michael J. Hayford
"""

from rayoptics.optical.model_constants import ht, slp


def __decode_lens__(lens_package):
    """ decode the lens_package tuple into its constituents

        Args:
            lens_package: a tuple or a |ParaxialModel|. If it's a tuple:

               - the first element is a |ParaxialModel|
               - the second element is a tuple with the begining and ending
                 indicies into lens model lists

        Returns:
            |ParaxialModel|, beginning and ending indicies into paraxial model
    """
    if type(lens_package) is tuple:
        parax_model = lens_package[0]
        bgn, end = lens_package[1]
    else:
        parax_model = lens_package
        bgn, end = (1, -1)
    return parax_model, bgn, end


def _separation_ratio(lens_package):
    """ the ratio of the backfocus to the mirror separation

        Args:
            see description in __decode_lens__
    """
    parax_model, bgn, end = __decode_lens__(lens_package)
    ax_ray = parax_model.ax
#    a = ax_ray[ht][end]/ax_ray[slp][end]
#    B = (ax_ray[ht][end] - ax_ray[ht][bgn])/ax_ray[slp][bgn]
#    s = a/B
    m = ax_ray[bgn][slp]/ax_ray[end][slp]
    s = m*ax_ray[end-1][ht]/(ax_ray[bgn][ht] - ax_ray[end-1][ht])
    return s


def _mag(lens_package):
    """ the magnification of the input paraxial lens over the specified range

        Args:
            see description in __decode_lens__
    """
    parax_model, bgn, end = __decode_lens__(lens_package)
    ax_ray = parax_model.ax
    m = ax_ray[bgn][slp]/ax_ray[end][slp]
    return m


def cassegrain(lens_package):
    """ calculate the conic constants for a cassegrain telescope

        Args:
            lens_package: a tuple or a |ParaxialModel|. If it's a tuple:

               - the first element is a |ParaxialModel|
               - the second element is a tuple with the begining and ending
                 indicies into lens model lists

        Returns:
            the conic constants of the primary and secondary mirrors
    """
    m = _mag(lens_package)
    k_pri = -1.0
    k_sec = -4.0*m/(m - 1.0)**2 - 1.0
    return k_pri, k_sec


def dall_kirkham(lens_package):
    """ calculate the conic constants for a dall-kirkham telescope

        Args:
            lens_package: a tuple or a |ParaxialModel|. If it's a tuple:

               - the first element is a |ParaxialModel|
               - the second element is a tuple with the begining and ending
                 indicies into lens model lists

        Returns:
            the conic constants of the primary and secondary mirrors
    """
    m = _mag(lens_package)
    s = _separation_ratio(lens_package)
    k_pri = (s*(m - 1.)*(m + 1.)**2)/((m + s)*m**3) - 1.
    k_sec = 0.0
    return k_pri, k_sec


def ritchey_chretien(lens_package):
    """ calculate the conic constants for a ritchey-chretien telescope

        Args:
            lens_package: a tuple or a |ParaxialModel|. If it's a tuple:

               - the first element is a |ParaxialModel|
               - the second element is a tuple with the begining and ending
                 indicies into lens model lists

        Returns:
            the conic constants of the primary and secondary mirrors
    """
    m = _mag(lens_package)
    s = _separation_ratio(lens_package)
    k_pri = -2.0*s/m**3 - 1.0
    k_sec = -(4.0*m*(m - 1.0) + 2.0*(m + s))/(m - 1.0)**3 - 1.0
    return k_pri, k_sec


def spheres(lens_package):
    """ function to revert the conic constants to spherical surfaces

        Args:
            lens_package: a tuple or a |ParaxialModel|. If it's a tuple:

               - the first element is a |ParaxialModel|
               - the second element is a tuple with the begining and ending
                 indicies into lens model lists

        Returns:
            the conic constants of the primary and secondary mirrors
    """
    k_pri = 0.0
    k_sec = 0.0
    return k_pri, k_sec



import unittest
from rayoptics.codev import reader


class TestReadSeqBuffer(unittest.TestCase):
    def test_continuation_proc(self):
        inpt = ['ab&', 'cd&', 'ef', 'gh']
        outpt = reader.read_seq_buffer(inpt)
        self.assertEqual(outpt, [['abcdef'], ['gh']])

    def test_eol_comment_precedence(self):
        inpt = ['ab; cd!ef; gh']
        outpt = reader.read_seq_buffer(inpt)
        self.assertEqual(outpt, [['ab'], ['cd']])

    def test_end_of_cmd(self):
        inpt = ['ab; cd; ef', 'gh;ij& ', 'kl; mn']
        outpt = reader.read_seq_buffer(inpt)
        self.assertEqual(outpt, [['ab'], ['cd'], ['ef'], ['gh'], ['ijkl'],
                                 ['mn']])

    def test_spaces_as_delimiters(self):
        inpt = ['s &', '.07 10 ', 's & ', ' 0 .001']
        outpt = reader.read_seq_buffer(inpt)
        self.assertEqual(outpt, [['s', '.07', '10'], ['s', '0', '.001']])

    def test_strip_comment_lines(self):
        inpt = ['tit "this is a title"', ' ! this is a comment line', 'dim m',
                'so 0 1e11 ! infinite object']
        outpt = reader.read_seq_buffer(inpt)
        self.assertEqual(outpt, [['tit', 'this is a title'], ['dim', 'm'],
                                 ['so', '0', '1e11']])

    def test_empty_buffer(self):
        inpt = []
        outpt = reader.read_seq_buffer(inpt)
        self.assertEqual(outpt, [])

    def test_empty_cmd_strings(self):
        inpt = ['', '! this comment will be stripped out', '', '    ! so will this one']
        outpt = reader.read_seq_buffer(inpt)
        self.assertEqual(outpt, [])


class TestNextLine(unittest.TestCase):
    def test_next_line_with_cont(self):
        inpt = ['ab&', 'cd&', 'ef']
        it = iter(inpt)
        outpt = reader.next_line(it)
        self.assertEqual(outpt, 'abcdef')

    def test_next_line(self):
        inpt = ['ab&', 'cd', 'ef']
        it = iter(inpt)
        outpt = reader.next_line(it)
        self.assertEqual(outpt, 'abcd')
        outpt = reader.next_line(it)
        self.assertEqual(outpt, 'ef')
        with self.assertRaises(StopIteration):
            outpt = reader.next_line(it)


class TestStripComments(unittest.TestCase):
    def test_strip_comment_line(self):
        inpt = '!abcd'
        outpt = reader.strip_comments(inpt)
        self.assertEqual(outpt, '')
        inpt = ' !abcd'
        outpt = reader.strip_comments(inpt)
        self.assertEqual(outpt, ' ')

    def test_strip_eol_comment(self):
        inpt = 'ab!cd'
        outpt = reader.strip_comments(inpt)
        self.assertEqual(outpt, 'ab')

    def test_strip_comment_no_comment(self):
        inpt = 'abcd'
        outpt = reader.strip_comments(inpt)
        self.assertEqual(outpt, 'abcd')


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2017 Michael J. Hayford
""" Module for different surface profile shapes

.. Created on Tue Aug  1 13:18:57 2017

.. codeauthor: Michael J. Hayford
"""

import numpy as np
from math import sqrt, copysign, sin, atan2
from rayoptics.util.misc_math import normalize
from .traceerror import TraceError, TraceMissedSurfaceError


def resize_list(lst, new_length, null_item=None):
    return lst + [null_item for item in range(new_length - len(lst))]


def intersect_parabola(cv, p, d, z_dir=1.0):
    ''' Intersect a parabolid, starting from an arbitrary point.

    Args:
        cv: vertex curvature
        p:  start point of the ray in the profile's coordinate system
        d:  direction cosine of the ray in the profile's coordinate system
        z_dir: +1 if propagation positive direction, -1 if otherwise
    '''
    # Intersection with a conic, starting from an arbitrary point.
    #
    # For quadratic equation ax**2 + bx + c = 0:
    #  ax2 = 2a
    #  cx2 = 2c
    ax2 = cv*(1. - d[2]*d[2])
    cx2 = cv*(p[0]*p[0] + p[1]*p[1]) - 2.0*p[2]
    b = cv*(d[0]*p[0] + d[1]*p[1]) - d[2]
    # Use z_dir to pick correct root
    s = cx2/(z_dir*sqrt(b*b - ax2*cx2) - b)

    p1 = p + s*d
    return s, p1


class SurfaceProfile:
    def __repr__(self):
        return "{!s}()".format(type(self).__name__)

    def update(self):
        pass

    def f(self, p):
        pass

    def normal(self, p):
        pass

    def sag(self, x, y):
        pass

    def profile(self, sd, dir=1, steps=6):
        pass

    def apply_scale_factor(self, scale_factor):
        pass

    def intersect(self, p0, d, eps, z_dir):
        ''' Intersect a profile, starting from an arbitrary point.

        Args:
            p0:  start point of the ray in the profile's coordinate system
            d:  direction cosine of the ray in the profile's coordinate system
            z_dir: +1 if propagation positive direction, -1 if otherwise
            eps: numeric tolerance for convergence of any iterative procedure

        Returns:
            tuple: distance to intersection point **s1**, intersection point **p**

        Raises:
            TraceMissedSurfaceError
        '''
        p = p1 = p0
        s1 = -self.f(p1)/np.dot(d, self.normal(p1))
        delta = abs(s1)
#        print("intersect", s1)
        while delta > eps:
            p = p1 + s1*d
            s2 = s1 - self.f(p)/np.dot(d, self.normal(p))
            delta = abs(s2 - s1)
#            print("intersect", s1, s2, delta)
            s1 = s2
        return s1, p


class Spherical(SurfaceProfile):
    """ Spherical surface profile parameterized by curvature. """

    def __init__(self, c=0.0, r=None):
        """ initialize a Spherical profile.

        Args:
            c: curvature
            r: radius of curvature. If zero, taken as planar. If r is
                specified, it overrides any input for c (curvature).
        """
        if r:
            self.r = r
        else:
            self.cv = c

    @property
    def r(self):
        if self.cv != 0.0:
            return 1.0/self.cv
        else:
            return 0.0

    @r.setter
    def r(self, radius):
        if radius != 0.0:
            self.cv = 1.0/radius
        else:
            self.cv = 0.0

    def __str__(self):
        return type(self).__name__ + " " + str(self.cv)

    def __repr__(self):
        return "{!s}(c={})".format(type(self).__name__, self.cv)

    def copyFrom(self, other):
        dispatch[type(self), type(other)](self, other)

    def copyDataFrom(self, other):
        self.cv = other.cv

    def mutate(self, new_profile):
        Spherical.copyDataFrom(new_profile, self)

    def apply_scale_factor(self, scale_factor):
        self.cv /= scale_factor

    def normal(self, p):
        return normalize(np.array(
                [-self.cv*p[0], -self.cv*p[1], 1.0-self.cv*p[2]]))

    def intersect_tangent_plane(self, p, d, eps, z_dir):
        # Welford's intersection with a sphere, starting from the tangent plane
        # transfer p to tangent plane of surface
        s0 = -p[2]/d[2]
        p0 = np.array([p[0] + s0*d[0], p[1] + s0*d[1], 0.])

        # Welford's 4.8, 4.9 and 4.12
        F = self.cv*(p0[0]*p0[0] + p0[1]*p0[1])
        G = d[2] - self.cv*(d[0]*p0[0] + d[1]*p0[1])
        try:
            s1 = F/(G + sqrt(G*G - F*self.cv))
        except ValueError:
            raise TraceMissedSurfaceError

        s = s0 + s1

        p1 = p + s*d
        return s, p1

    def intersect(self, p, d, eps, z_dir):
        ''' Intersection with a sphere, starting from an arbitrary point. '''

        # Substitute expressions equivalent to Welford's 4.8 and 4.9
        # For quadratic equation ax**2 + bx + c = 0:
        #  ax2 = 2a
        #  cx2 = 2c
        ax2 = self.cv
        cx2 = self.cv * p.dot(p) - 2*p[2]
        b = self.cv * d.dot(p) - d[2]
        try:
            # Use z_dir to pick correct root
            s = cx2/(z_dir*sqrt(b*b - ax2*cx2) - b)
        except ValueError:
            raise TraceMissedSurfaceError

        p1 = p + s*d
        return s, p1

    def f(self, p):
        return p[2] - 0.5*self.cv*(np.dot(p, p))

    def sag(self, x, y):
        if self.cv != 0.0:
            r = 1/self.cv
            try:
                adj = sqrt(r*r - x*x - y*y)
            except ValueError:
                raise TraceMissedSurfaceError(self, (x, y))
            else:
                return r*(1 - abs(adj/r))
        else:
            return 0

    def profile(self, sd, dir=1, steps=6):
        prf = []
        if len(sd) == 1:
            sd_lwr = -sd[0]
            sd_upr = sd[0]
        else:
            sd_lwr = sd[0]
            sd_upr = sd[1]

        if self.cv != 0.0:
            r = 1/self.cv
            try:
                adj_upr = sqrt(r*r - sd_upr*sd_upr)
            except ValueError:
                sd_upr = abs(r)
                adj_upr = 0.
            ang_upr = atan2(sd_upr, adj_upr)
            try:
                adj_lwr = sqrt(r*r - sd_lwr*sd_lwr)
            except ValueError:
                sd_lwr = -abs(r)
                adj_lwr = 0.
            ang_lwr = atan2(sd_lwr, adj_lwr)
            ang = 0.5*(ang_upr - ang_lwr)
            da = dir*copysign(ang/steps, self.cv)
            sa = sin(da)
            ca = sqrt(1.0 - sa*sa)
            sab = sb = -dir*(sd_upr/r)
            cab = cb = abs(adj_upr/r)
            for i in range(2*steps):
                prf.append([r*(1-cab), r*sab])
                # print(i, r*(1-cab), r*sab)
                sab = sa*cb + sb*ca
                cab = ca*cb - sa*sb
                sb = sab
                cb = cab
            prf.append([r*(1-cab), r*sab])
            # print(2*steps, r*(1-cab), r*sab)
        else:
            prf.append([0, dir*sd_lwr])
            prf.append([0, dir*sd_upr])
        return prf


class Conic(SurfaceProfile):
    """ Conic surface profile parameterized by curvature and conic constant.

    Conics produced for conic constant values:
        cc > 0.0: oblate spheroid
        cc = 0.0: sphere
        cc < 0.0 and > -1.0: ellipsoid
        cc = -1.0: paraboloid
        cc < -1.0: hyperboloid
    """
    def __init__(self, c=0.0, cc=0.0, r=None, ec=None):
        """ initialize a Conic profile.

        Args:
            c: curvature
            r: radius of curvature. If zero, taken as planar. If r is
                specified, it overrides any input for c (curvature).
            cc: conic constant
            ec: conic asphere (= cc + 1). If ec is specified, it overrides any
                input for the conic constant (cc).
        """
        if r:
            self.r = r
        else:
            self.cv = c

        if ec:
            self.ec = ec
        else:
            self.cc = cc

    @property
    def r(self):
        if self.cv != 0.0:
            return 1.0/self.cv
        else:
            return 0.0

    @r.setter
    def r(self, radius):
        if radius != 0.0:
            self.cv = 1.0/radius
        else:
            self.cv = 0.0

    @property
    def ec(self):
        return self.cc + 1.0

    @ec.setter
    def ec(self, ec):
        self.cc = ec - 1.0

    def __str__(self):
        return type(self).__name__ + " " + str(self.cv) + " " + \
                                           str(self.cc)

    def __repr__(self):
        return "{!s}(c={}, cc={})".format(type(self).__name__,
                                          self.cv, self.cc)

    def copyFrom(self, other):
        dispatch[type(self), type(other)](self, other)

    def copyDataFrom(self, other):
        self.cv = other.cv
        self.cc = other.cc

    def apply_scale_factor(self, scale_factor):
        self.cv /= scale_factor

    def normal(self, p):
        return normalize(np.array(
                [-self.cv*p[0],
                 -self.cv*p[1],
                 1.0-(self.cc+1.0)*self.cv*p[2]]))

    def intersect_tangent_plane(self, p, d, eps, z_dir):
        # Welford's intersection with a conic, starting from the tangent plane
        # transfer p to tangent plane of surface
        s0 = -p[2]/d[2]
        p0 = np.array([p[0] + s0*d[0], p[1] + s0*d[1], 0.])

        # Welford's 4.8, 4.9 and 4.28
        ax2 = self.cv*(1. + self.cc*d[2]*d[2])
        F = self.cv*(p0[0]*p0[0] + p0[1]*p0[1])
        G = d[2] - self.cv*(d[0]*p0[0] + d[1]*p0[1])
        try:
            s1 = F/(G + z_dir*sqrt(G*G - F*ax2))
        except ValueError:
            raise TraceMissedSurfaceError

        s = s0 + s1
        p1 = p + s*d
        return s, p1

    def intersect(self, p, d, eps, z_dir):
        ''' Intersection with a conic, starting from an arbitrary point.'''

        # For quadratic equation ax**2 + bx + c = 0:
        #  ax2 = 2a
        #  cx2 = 2c
        ax2 = self.cv*(1. + self.cc*d[2]*d[2])
        cx2 = self.cv*(p[0]*p[0] + p[1]*p[1] + self.ec*p[2]*p[2]) - 2.0*p[2]
        b = self.cv*(d[0]*p[0] + d[1]*p[1] + self.ec*d[2]*p[2]) - d[2]
        try:
            # Use z_dir to pick correct root
            s = cx2/(z_dir*sqrt(b*b - ax2*cx2) - b)
        except ValueError:
            raise TraceMissedSurfaceError

        p1 = p + s*d
        return s, p1

    def f(self, p):
        return p[2] - 0.5*self.cv*(p[0]*p[0] +
                                   p[1]*p[1] +
                                   (self.cc+1.0)*p[2]*p[2])

    def sag(self, x, y):
        r2 = x*x + y*y
        try:
            z = self.cv*r2/(1. + sqrt(1. - (self.cc+1.0)*self.cv*self.cv*r2))
        except ValueError:
            raise TraceMissedSurfaceError
        return z

    def profile(self, sd, dir=1, steps=6):
        prf = []
        if len(sd) == 1:
            sd_lwr = -sd[0]
            sd_upr = sd[0]
        else:
            sd_lwr = sd[0]
            sd_upr = sd[1]

        if self.cv != 0.0:
            delta = dir*(sd_upr-sd_lwr)/(2*steps)
            y = sd_lwr if dir > 0 else sd_upr
            z = self.sag(0., y)
            prf.append([z, y])
            for i in range(2*steps):
                y += delta
                z = self.sag(0., y)
                prf.append([z, y])
                # print(i, z, y)
        else:
            prf.append([0, dir*sd_lwr])
            prf.append([0, dir*sd_upr])
        return prf


def append_pt_to_2d_profile(surface_profile, y, poly_profile):
    """ calc surface sag at y and append to poly if ok, else return None """
    try:
        z = surface_profile.sag(0, y)
    except TraceError:
        return None
    else:
        poly_profile.append([z, y])
        return z


def aspheric_profile(surface_profile, sd, dir=1, steps=21):
    if steps < 21:
        steps = 21

    prf = []
    if len(sd) == 1:
        sd_lwr = -sd[0]
        sd_upr = sd[0]
    else:
        sd_lwr = sd[0]
        sd_upr = sd[1]

    if surface_profile.max_nonzero_coef > 0 or surface_profile.cv != 0.0:
        delta = dir*(sd_upr-sd_lwr)/(2*steps)
        y = sd_lwr if dir > 0 else sd_upr
        append_pt_to_2d_profile(surface_profile, y, prf)
        for i in range(2*steps):
            y += delta
            append_pt_to_2d_profile(surface_profile, y, prf)
    else:
        prf.append([0, dir*sd_lwr])
        prf.append([0, dir*sd_upr])
    return prf


class EvenPolynomial(SurfaceProfile):
    """ Even Polynomial asphere up to 20th order, on base conic. """
    def __init__(self, c=0.0, cc=0.0, r=None, ec=None, coefs=None):
        """ initialize a EvenPolynomial profile.

        Args:
            c: curvature
            r: radius of curvature. If zero, taken as planar. If r is
                specified, it overrides any input for c (curvature).
            cc: conic constant
            ec: conic asphere (= cc + 1). If ec is specified, it overrides any
                input for the conic constant (cc).
            coefs: a list of even power coefficents, starting with the
                quadratic term, and not exceeding the 20th order term.
        """
        if r:
            self.r = r
        else:
            self.cv = c

        if ec:
            self.ec = ec
        else:
            self.cc = cc

        if coefs:
            self.coefs = coefs
        else:
            self.coefs = []
        self.max_nonzero_coef = 0
        self.coef2 = 0.0
        self.coef4 = 0.0
        self.coef6 = 0.0
        self.coef8 = 0.0
        self.coef10 = 0.0
        self.coef12 = 0.0
        self.coef14 = 0.0
        self.coef16 = 0.0
        self.coef18 = 0.0
        self.coef20 = 0.0

    @property
    def r(self):
        if self.cv != 0.0:
            return 1.0/self.cv
        else:
            return 0.0

    @r.setter
    def r(self, radius):
        if radius != 0.0:
            self.cv = 1.0/radius
        else:
            self.cv = 0.0

    @property
    def ec(self):
        return self.cc + 1.0

    @ec.setter
    def ec(self, ec):
        self.cc = ec - 1.0

    def __str__(self):
        return type(self).__name__ + " " + str(self.cv) + " " + \
                                           str(self.cc)

    def __repr__(self):
        return (type(self).__name__ + '(c=' + repr(self.cv) +
                ', cc=' + repr(self.cc) +
                ', coefs=' + repr(self.coefs) + ')')

    def copyFrom(self, other):
        dispatch[type(self), type(other)](self, other)

    def copyDataFrom(self, other):
        self.cv = other.cv
        self.cc = other.cc
        self.coef2 = other.coef2
        self.coef4 = other.coef4
        self.coef6 = other.coef6
        self.coef8 = other.coef8
        self.coef10 = other.coef10
        self.coef12 = other.coef12
        self.coef14 = other.coef14
        self.coef16 = other.coef16
        self.coef18 = other.coef18
        self.coef20 = other.coef20

    def gen_coef_list(self):
        if len(self.coefs) == 0:
            self.coefs = []
            self.coefs.append(self.coef2)
            self.coefs.append(self.coef4)
            self.coefs.append(self.coef6)
            self.coefs.append(self.coef8)
            self.coefs.append(self.coef10)
            self.coefs.append(self.coef12)
            self.coefs.append(self.coef14)
            self.coefs.append(self.coef16)
            self.coefs.append(self.coef18)
            self.coefs.append(self.coef20)
            self.max_nonzero_coef = -1
        for i, c in enumerate(self.coefs):
            if c != 0.0:
                self.max_nonzero_coef = i
        self.max_nonzero_coef += 1

    def apply_scale_factor(self, scale_factor):
        self.cv /= scale_factor
        sf_sqr = scale_factor**2
        self.coef2 *= sf_sqr
        self.coef4 *= sf_sqr**2
        self.coef6 *= sf_sqr**3
        self.coef8 *= sf_sqr**4
        self.coef10 *= sf_sqr**5
        self.coef12 *= sf_sqr**6
        self.coef14 *= sf_sqr**7
        self.coef16 *= sf_sqr**8
        self.coef18 *= sf_sqr**9
        self.coef20 *= sf_sqr**10

    def update(self):
        self.gen_coef_list()

    def normal(self, p):
        # sphere + conic contribution
        r2 = p[0]*p[0] + p[1]*p[1]
        e = self.cv/sqrt(1. - (self.cc+1.0)*self.cv*self.cv*r2)

        # polynomial asphere contribution
        e_asp = 0.0
        r_pow = 1.0
        c_coef = 2.0
        for i in range(self.max_nonzero_coef):
            e_asp += c_coef*self.coefs[i]*r_pow
            c_coef += 2.0
            r_pow *= r2

        e_tot = e + e_asp
        return normalize(np.array([-e_tot*p[0], -e_tot*p[1], 1.0]))

    def sag(self, x, y):
        r2 = x*x + y*y
        try:
            # sphere + conic contribution
            z = self.cv*r2/(1. + sqrt(1. - (self.cc+1.0)*self.cv*self.cv*r2))
        except ValueError:
            raise TraceMissedSurfaceError

        # polynomial asphere contribution
        z_asp = 0.0
        r_pow = r2
        for i in range(self.max_nonzero_coef):
            z_asp += self.coefs[i]*r_pow
            r_pow *= r2

        z_tot = z + z_asp
        return z_tot

    def f(self, p):
        return p[2] - self.sag(p[0], p[1])

    def profile(self, sd, dir=1, steps=21):
        return aspheric_profile(self, sd, dir, steps)


class RadialPolynomial(SurfaceProfile):
    """ Radial Polynomial asphere, on base conic.

    Conics produced for conic asphere values:
        ec > 1.0: oblate spheroid
        ec = 1.0: sphere
        ec > 0.0 and < 1.0: ellipsoid
        ec = 0.0: paraboloid
        ec < 0.0: hyperboloid
    """
    initial_size = 10

    def __init__(self, c=0.0, cc=None, r=None, ec=1.0, coefs=None):
        """ initialize a RadialPolynomial profile.

        Args:
            c: curvature
            r: radius of curvature. If zero, taken as planar. If r is
                specified, it overrides any input for c (curvature).
            ec: conic asphere.
            cc: conic constant (= ec - 1). If cc is specified, it overrides any
                input for the conic asphere (ec).
            coefs: a list of radial coefficents, starting with the
                constant term, (and not exceeding the 10th order term).
        """
        if r:
            self.r = r
        else:
            self.cv = c

        if cc:
            self.cc = cc
        else:
            self.ec = ec

        if coefs:
            self.coefs = coefs
        else:
            self.coefs = [0. for i in range(self.initial_size)]
        self.max_nonzero_coef = 0
        self.coef1 = 0.0
        self.coef2 = 0.0
        self.coef3 = 0.0
        self.coef4 = 0.0
        self.coef5 = 0.0
        self.coef6 = 0.0
        self.coef7 = 0.0
        self.coef8 = 0.0
        self.coef9 = 0.0
        self.coef10 = 0.0

    @property
    def r(self):
        if self.cv != 0.0:
            return 1.0/self.cv
        else:
            return 0.0

    @r.setter
    def r(self, radius):
        if radius != 0.0:
            self.cv = 1.0/radius
        else:
            self.cv = 0.0

    @property
    def cc(self):
        return self.ec - 1.0

    @cc.setter
    def cc(self, cc):
        self.ec = cc + 1.0

#    @property
    def get_coef(self, exp):
        return self.coefs[exp]

#    @coef.setter
    def set_coef(self, exp, value):
        try:
            self.coefs[exp] = value
        except IndexError:
            self.coefs = resize_list(self.coefs, exp, null_item=0.0)
            self.coefs[exp] = value

    def __str__(self):
        return type(self).__name__ + " " + str(self.cv) + " " + \
                                           str(self.ec)

    def __repr__(self):
        return (type(self).__name__ + '(c=' + repr(self.cv) +
                ', ec=' + repr(self.ec) +
                ', coefs=' + repr(self.coefs) + ')')

    def copyFrom(self, other):
        dispatch[type(self), type(other)](self, other)

    def copyDataFrom(self, other):
        self.cv = other.cv
        self.ec = other.ec
        self.coefs = other.coefs.copy()
        self.coefs = other.coefs
        self.coef1 = other.coef1
        self.coef2 = other.coef2
        self.coef3 = other.coef3
        self.coef4 = other.coef4
        self.coef5 = other.coef5
        self.coef6 = other.coef6
        self.coef7 = other.coef7
        self.coef8 = other.coef8
        self.coef9 = other.coef9
        self.coef10 = other.coef10

    def gen_coef_list(self):
        if len(self.coefs) == 0:
            self.coefs = []
            self.coefs.append(self.coef1)
            self.coefs.append(self.coef2)
            self.coefs.append(self.coef3)
            self.coefs.append(self.coef4)
            self.coefs.append(self.coef5)
            self.coefs.append(self.coef6)
            self.coefs.append(self.coef7)
            self.coefs.append(self.coef8)
            self.coefs.append(self.coef9)
            self.coefs.append(self.coef10)
            self.max_nonzero_coef = -1
        for i, c in enumerate(self.coefs):
            if c != 0.0:
                self.max_nonzero_coef = i
        self.max_nonzero_coef += 1

    def apply_scale_factor(self, scale_factor):
        self.cv /= scale_factor
        self.coef1 *= scale_factor
        self.coef2 *= scale_factor**2
        self.coef3 *= scale_factor**3
        self.coef4 *= scale_factor**4
        self.coef5 *= scale_factor**5
        self.coef6 *= scale_factor**6
        self.coef7 *= scale_factor**7
        self.coef8 *= scale_factor**8
        self.coef9 *= scale_factor**9
        self.coef10 *= scale_factor**10

    def update(self):
        self.gen_coef_list()

    def normal(self, p):
        # sphere + conic contribution
        r2 = p[0]*p[0] + p[1]*p[1]
        r = sqrt(r2)
        e = self.cv/sqrt(1. - self.ec*self.cv*self.cv*r2)

        # polynomial asphere contribution - compute using Horner's Rule
        e_asp = 0.0
        if r == 0.0:
            r_pow = 1.0
        else:
            # Initialize to 1/r because we multiply by r's components p[0] and
            # p[1] at the final normalization step.
            r_pow = 1.0/r
        c_coef = 1.0
        for coef in self.coefs[1:self.max_nonzero_coef]:
            e_asp += c_coef*coef*r_pow
            c_coef += 1.0
            r_pow *= r

        e_tot = e + e_asp
        return normalize(np.array([-e_tot*p[0], -e_tot*p[1], 1.0]))

    def sag(self, x, y):
        r2 = x*x + y*y
        r = sqrt(r2)
        try:
            # sphere + conic contribution
            z = self.cv*r2/(1. + sqrt(1. - self.ec*self.cv*self.cv*r2))
        except ValueError:
            raise TraceMissedSurfaceError

        # polynomial asphere contribution
        z_asp = 0.0
        r_pow = 1.0
        for coef in self.coefs[:self.max_nonzero_coef]:
            z_asp += coef*r_pow
            r_pow *= r

        z_tot = z + z_asp
        return z_tot

    def f(self, p):
        return p[2] - self.sag(p[0], p[1])

    def profile(self, sd, dir=1, steps=21):
        return aspheric_profile(self, sd, dir, steps)


dispatch = {
  (Spherical, Spherical): Spherical.copyDataFrom,
  (Spherical, Conic): Spherical.copyDataFrom,
  (Spherical, EvenPolynomial): Spherical.copyDataFrom,
  (Spherical, RadialPolynomial): Spherical.copyDataFrom,
  (Conic, Spherical): Spherical.copyDataFrom,
  (Conic, Conic): Conic.copyDataFrom,
  (Conic, EvenPolynomial): Conic.copyDataFrom,
  (Conic, RadialPolynomial): Conic.copyDataFrom,
  (EvenPolynomial, Spherical): Spherical.copyDataFrom,
  (EvenPolynomial, Conic): Conic.copyDataFrom,
  (EvenPolynomial, EvenPolynomial): EvenPolynomial.copyDataFrom,
  (EvenPolynomial, RadialPolynomial): EvenPolynomial.copyDataFrom,
  (RadialPolynomial, Spherical): Spherical.copyDataFrom,
  (RadialPolynomial, Conic): Conic.copyDataFrom,
  (RadialPolynomial, EvenPolynomial): EvenPolynomial.copyDataFrom,
  (RadialPolynomial, RadialPolynomial): RadialPolynomial.copyDataFrom,
}


def mutate_profile(old_profile, new_profile_type):
    new_profile = eval(new_profile_type+'()')
    new_profile.copyFrom(old_profile)
    return new_profile


def test():
    s1 = Spherical(0.0)
    s2 = Spherical(0.1)

    dir0 = np.array([0., 0., 1.])
    v1 = normalize(np.array([1., 1., 1.]))
    p0 = np.array([0., 0., -1.])
    p1 = np.array([0., 1., -1.])

    p0s1 = s1.intersect(p0, dir0)
    print(p0s1)
    p1s1 = s1.intersect(p1, dir0)
    print(p1s1)

    p0s2 = s2.intersect(p0, dir0)
    print(p0s2)
    p1s2 = s2.intersect(p1, dir0)
    print(p1s2)
    dir_p1s2 = s2.normal(p1s2[1])
    print(dir_p1s2)

    print("pass")


if __name__ == "__main__":
    test()
